from keras import layers
from keras import ops

from keras_hub.src.models.depth_anything.interpolate import interpolate
from keras_hub.src.utils.keras_utils import standardize_data_format


class DepthAnythingTokenToImage(layers.Layer):
    """A layer that converts tokens into images.

    Args:
        hidden_dim: int. The number of units in the hidden layers.
        patch_height: int. The height of each patch.
        patch_width: int. The width of each patch.
        num_cls_tokens: int. The number of class tokens at the beginning of
            the sequence. Defaults to `1`.
        num_register_tokens: int. The number of register tokens after the
            class tokens. Defaults to `0`.
        data_format: `None` or str. If specified, either `"channels_last"` or
            `"channels_first"`. The ordering of the dimensions in the
            inputs. `"channels_last"` corresponds to inputs with shape
            `(batch_size, height, width, channels)`
            while `"channels_first"` corresponds to inputs with shape
            `(batch_size, channels, height, width)`. It defaults to the
            `image_data_format` value found in your Keras config file at
            `~/.keras/keras.json`. If you never set it, then it will be
            `"channels_last"`.
        **kwargs: other keyword arguments passed to `keras.layers.Layer`,
            including `name`, `dtype` etc.
    """

    def __init__(
        self,
        hidden_dim,
        patch_height,
        patch_width,
        num_cls_tokens=1,
        num_register_tokens=0,
        data_format=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.hidden_dim = int(hidden_dim)
        self.patch_height = int(patch_height)
        self.patch_width = int(patch_width)
        self.num_cls_tokens = int(num_cls_tokens)
        self.num_register_tokens = int(num_register_tokens)
        self.data_format = standardize_data_format(data_format)
        # Always use channels_last for reshaping first.
        self.target_shape = (
            self.patch_height,
            self.patch_width,
            self.hidden_dim,
        )

    def call(self, inputs):
        # Remove the cls token.
        x = inputs[:, self.num_cls_tokens + self.num_register_tokens :, ...]

        x = ops.reshape(x, (ops.shape(x)[0],) + self.target_shape)
        if self.data_format == "channels_first":
            x = ops.transpose(x, (0, 3, 1, 2))
        return x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "hidden_dim": self.hidden_dim,
                "patch_height": self.patch_height,
                "patch_width": self.patch_width,
                "num_cls_tokens": self.num_cls_tokens,
                "num_register_tokens": self.num_register_tokens,
            }
        )
        return config

    def compute_output_shape(self, input_shape):
        output_shape = [input_shape[0], *self.target_shape]
        if self.data_format == "channels_first":
            output_shape = [
                output_shape[0],
                output_shape[3],
                output_shape[1],
                output_shape[2],
            ]
        return output_shape


class DepthAnythingReassembleLayer(layers.Layer):
    """A layer that resizes the input images.

    Args:
        hidden_dim: int. The number of units in the hidden layers.
        factor: float. The resizing factor. If `factor > 1`, the layer upsamples
            the input. If `factor < 1`, the layer downsamples the input. If
            `factor == 1`, the layer only applies a linear projection.
        data_format: `None` or str. If specified, either `"channels_last"` or
            `"channels_first"`. The ordering of the dimensions in the
            inputs. `"channels_last"` corresponds to inputs with shape
            `(batch_size, height, width, channels)`
            while `"channels_first"` corresponds to inputs with shape
            `(batch_size, channels, height, width)`. It defaults to the
            `image_data_format` value found in your Keras config file at
            `~/.keras/keras.json`. If you never set it, then it will be
            `"channels_last"`.
        **kwargs: other keyword arguments passed to `keras.layers.Layer`,
            including `name`, `dtype` etc.
    """

    def __init__(self, hidden_dim, factor, data_format=None, **kwargs):
        super().__init__(**kwargs)
        self.hidden_dim = int(hidden_dim)
        self.factor = float(factor)
        self.data_format = standardize_data_format(data_format)

        self.projection = layers.Conv2D(
            filters=self.hidden_dim,
            kernel_size=1,
            data_format=self.data_format,
            use_bias=True,
            dtype=self.dtype_policy,
            name="projection",
        )
        if self.factor > 1:
            self.padding = layers.Identity(
                dtype=self.dtype_policy, name="padding"
            )
            self.resize = layers.Conv2DTranspose(
                filters=self.hidden_dim,
                kernel_size=int(self.factor),
                strides=int(self.factor),
                data_format=self.data_format,
                use_bias=True,
                dtype=self.dtype_policy,
                name="resize",
            )
        elif self.factor == 1:
            self.padding = layers.Identity(
                dtype=self.dtype_policy, name="padding"
            )
            self.resize = layers.Identity(
                dtype=self.dtype_policy, name="resize"
            )
        elif self.factor < 1:
            self.padding = layers.ZeroPadding2D(
                padding=(1, 1),
                data_format=self.data_format,
                dtype=self.dtype_policy,
                name="padding",
            )
            self.resize = layers.Conv2D(
                filters=self.hidden_dim,
                kernel_size=3,
                strides=int(1 / self.factor),
                data_format=self.data_format,
                use_bias=True,
                dtype=self.dtype_policy,
                name="resize",
            )

    def build(self, inputs_shape):
        self.projection.build(inputs_shape)
        inputs_shape = self.projection.compute_output_shape(inputs_shape)
        self.padding.build(inputs_shape)
        inputs_shape = self.padding.compute_output_shape(inputs_shape)
        self.resize.build(inputs_shape)

    def call(self, inputs, training=None):
        x = self.projection(inputs, training=training)
        x = self.padding(x, training=training)
        return self.resize(x, training=training)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "hidden_dim": self.hidden_dim,
                "factor": self.factor,
            }
        )
        return config

    def compute_output_shape(self, input_shape):
        output_shape = list(input_shape)
        if self.data_format == "channels_first":
            output_shape[1] = self.hidden_dim
            output_shape[2] = int(output_shape[2] * self.factor)
            output_shape[3] = int(output_shape[3] * self.factor)
        else:
            output_shape[1] = int(output_shape[1] * self.factor)
            output_shape[2] = int(output_shape[2] * self.factor)
            output_shape[3] = self.hidden_dim
        return output_shape


class DepthAnythingPreActResidualLayer(layers.Layer):
    """A ReLU + Conv2D layer.

    Args:
        hidden_dim: int. The number of units in the hidden layers.
        data_format: `None` or str. If specified, either `"channels_last"` or
            `"channels_first"`. The ordering of the dimensions in the
            inputs. `"channels_last"` corresponds to inputs with shape
            `(batch_size, height, width, channels)`
            while `"channels_first"` corresponds to inputs with shape
            `(batch_size, channels, height, width)`. It defaults to the
            `image_data_format` value found in your Keras config file at
            `~/.keras/keras.json`. If you never set it, then it will be
            `"channels_last"`.
        **kwargs: other keyword arguments passed to `keras.layers.Layer`,
            including `name`, `dtype` etc.
    """

    def __init__(self, hidden_dim, data_format=None, **kwargs):
        super().__init__(**kwargs)
        self.hidden_dim = int(hidden_dim)
        self.data_format = standardize_data_format(data_format)

        self.activation1 = layers.ReLU(
            dtype=self.dtype_policy, name="activation1"
        )
        self.padding1 = layers.ZeroPadding2D(
            padding=(1, 1),
            data_format=self.data_format,
            dtype=self.dtype_policy,
            name="padding1",
        )
        self.convolution1 = layers.Conv2D(
            filters=self.hidden_dim,
            kernel_size=3,
            strides=1,
            data_format=self.data_format,
            use_bias=True,
            dtype=self.dtype_policy,
            name="convolution1",
        )
        self.activation2 = layers.ReLU(
            dtype=self.dtype_policy, name="activation2"
        )
        self.padding2 = layers.ZeroPadding2D(
            padding=(1, 1),
            data_format=self.data_format,
            dtype=self.dtype_policy,
            name="padding2",
        )
        self.convolution2 = layers.Conv2D(
            filters=self.hidden_dim,
            kernel_size=3,
            strides=1,
            data_format=self.data_format,
            use_bias=True,
            dtype=self.dtype_policy,
            name="convolution2",
        )

    def build(self, inputs_shape):
        self.activation1.build(inputs_shape)
        self.padding1.build(inputs_shape)
        inputs_shape = self.padding1.compute_output_shape(inputs_shape)
        self.convolution1.build(inputs_shape)
        inputs_shape = self.convolution1.compute_output_shape(inputs_shape)
        self.activation2.build(inputs_shape)
        self.padding2.build(inputs_shape)
        inputs_shape = self.padding2.compute_output_shape(inputs_shape)
        self.convolution2.build(inputs_shape)

    def call(self, inputs, training=None):
        residual = inputs
        x = self.activation1(inputs, training=training)
        x = self.padding1(x, training=training)
        x = self.convolution1(x, training=training)
        x = self.activation2(x, training=training)
        x = self.padding2(x, training=training)
        x = self.convolution2(x, training=training)
        return ops.add(x, residual)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "hidden_dim": self.hidden_dim,
            }
        )
        return config

    def compute_output_shape(self, input_shape):
        return input_shape


class DepthAnythingFeatureFusionLayer(layers.Layer):
    """A layer that fuses the incoming features.

    Args:
        hidden_dim: int. The number of units in the hidden layers.
        size: tuple of int. The target size of the output feature map.
        data_format: `None` or str. If specified, either `"channels_last"` or
            `"channels_first"`. The ordering of the dimensions in the
            inputs. `"channels_last"` corresponds to inputs with shape
            `(batch_size, height, width, channels)`
            while `"channels_first"` corresponds to inputs with shape
            `(batch_size, channels, height, width)`. It defaults to the
            `image_data_format` value found in your Keras config file at
            `~/.keras/keras.json`. If you never set it, then it will be
            `"channels_last"`.
        **kwargs: other keyword arguments passed to `keras.layers.Layer`,
            including `name`, `dtype` etc.
    """

    def __init__(self, hidden_dim, size, data_format=None, **kwargs):
        super().__init__(**kwargs)
        self.hidden_dim = int(hidden_dim)
        self.size = tuple(int(s) for s in size)
        self.data_format = standardize_data_format(data_format)

        self.residual_layer1 = DepthAnythingPreActResidualLayer(
            hidden_dim=self.hidden_dim,
            data_format=self.data_format,
            dtype=self.dtype_policy,
            name="residual_layer1",
        )
        self.residual_layer2 = DepthAnythingPreActResidualLayer(
            hidden_dim=self.hidden_dim,
            data_format=self.data_format,
            dtype=self.dtype_policy,
            name="residual_layer2",
        )
        self.projection = layers.Conv2D(
            filters=self.hidden_dim,
            kernel_size=1,
            data_format=self.data_format,
            use_bias=True,
            dtype=self.dtype_policy,
            name="projection",
        )

    def build(self, inputs_shape):
        self.residual_layer1.build(inputs_shape)
        self.residual_layer2.build(inputs_shape)
        inputs_shape = list(inputs_shape)
        if self.data_format == "channels_last":
            inputs_shape[1] = self.size[0]
            inputs_shape[2] = self.size[1]
        else:
            inputs_shape[2] = self.size[0]
            inputs_shape[3] = self.size[1]
        self.projection.build(inputs_shape)

    def call(self, inputs, residual=None, training=None):
        if residual is not None:
            inputs = ops.add(
                inputs, self.residual_layer1(residual, training=training)
            )

        x = self.residual_layer2(inputs, training=training)
        x = interpolate(x, size=self.size, data_format=self.data_format)
        return self.projection(x, training=training)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "hidden_dim": self.hidden_dim,
                "size": self.size,
            }
        )
        return config

    def compute_output_shape(self, input_shape):
        input_shape = self.residual_layer2.compute_output_shape(input_shape)
        input_shape = list(input_shape)
        if self.data_format == "channels_last":
            input_shape[1] = self.size[0]
            input_shape[2] = self.size[1]
        else:
            input_shape[2] = self.size[0]
            input_shape[3] = self.size[1]
        return self.projection.compute_output_shape(input_shape)


class DepthAnythingNeck(layers.Layer):
    """A DepthAnything neck layer.

    Args:
        patch_size: int. The size of one side of each patch.
        image_size: tuple of ints. The (height, width) of the input images.
        backbone_hidden_dim: int. The number of units in the backbone layers.
        neck_hidden_dims: List of int. The number of units in each neck layer.
        reassemble_factors: List of float. The resizing factor in each neck
            layer.
        fusion_hidden_dim: int. The number of units in the fusion layers.
        num_cls_tokens: int. The number of class tokens at the beginning of
            the sequence. Defaults to `1`.
        num_register_tokens: int. The number of register tokens after the
            class tokens. Defaults to `0`.
        data_format: `None` or str. If specified, either `"channels_last"` or
            `"channels_first"`. The ordering of the dimensions in the
            inputs. `"channels_last"` corresponds to inputs with shape
            `(batch_size, height, width, channels)`
            while `"channels_first"` corresponds to inputs with shape
            `(batch_size, channels, height, width)`. It defaults to the
            `image_data_format` value found in your Keras config file at
            `~/.keras/keras.json`. If you never set it, then it will be
            `"channels_last"`.
        **kwargs: other keyword arguments passed to `keras.layers.Layer`,
            including `name`, `dtype` etc.
    """

    def __init__(
        self,
        patch_size,
        image_size,
        backbone_hidden_dim,
        neck_hidden_dims,
        reassemble_factors,
        fusion_hidden_dim,
        num_cls_tokens=1,
        num_register_tokens=0,
        data_format=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.patch_size = int(patch_size)
        self.image_size = (int(image_size[0]), int(image_size[1]))
        self.backbone_hidden_dim = int(backbone_hidden_dim)
        self.neck_hidden_dims = tuple(int(d) for d in neck_hidden_dims)
        self.reassemble_factors = tuple(float(f) for f in reassemble_factors)
        self.fusion_hidden_dim = int(fusion_hidden_dim)
        self.num_cls_tokens = int(num_cls_tokens)
        self.num_register_tokens = int(num_register_tokens)
        self.data_format = standardize_data_format(data_format)
        if len(self.neck_hidden_dims) != len(self.reassemble_factors):
            raise ValueError(
                "`DepthAnythingNeck` expects the length of `neck_hidden_dims` "
                "and `reassemble_factors` to be the same. "
                f"Received: neck_hidden_dims={neck_hidden_dims}, "
                f"reassemble_factors={reassemble_factors}"
            )

        # Calculate the patch sizes for token to image layers.
        patch_height = self.image_size[0] // self.patch_size
        patch_width = self.image_size[1] // self.patch_size
        # Calculate the sizes for fusion layers.
        fusion_sizes = [
            (int(patch_height * factor), int(patch_width * factor))
            for factor in reversed(self.reassemble_factors[:-1])
        ]
        fusion_sizes = fusion_sizes + [
            (fusion_sizes[-1][0] * 2, fusion_sizes[-1][1] * 2)
        ]

        self.token_to_images = [
            DepthAnythingTokenToImage(
                hidden_dim=backbone_hidden_dim,
                patch_height=patch_height,
                patch_width=patch_width,
                num_cls_tokens=num_cls_tokens,
                num_register_tokens=num_register_tokens,
                data_format=self.data_format,
                dtype=self.dtype_policy,
                name=f"token_to_images_{i}",
            )
            for i in range(len(self.neck_hidden_dims))
        ]
        self.reassemble_stage = [
            DepthAnythingReassembleLayer(
                hidden_dim=hidden_dim,
                factor=factor,
                data_format=self.data_format,
                dtype=self.dtype_policy,
                name=f"reassemble_stage_{i}",
            )
            for i, (hidden_dim, factor) in enumerate(
                zip(self.neck_hidden_dims, self.reassemble_factors)
            )
        ]
        self.paddings = [
            layers.ZeroPadding2D(
                padding=(1, 1),
                data_format=self.data_format,
                dtype=self.dtype_policy,
                name=f"paddings_{i}",
            )
            for i in range(len(self.neck_hidden_dims))
        ]
        self.convs = [
            layers.Conv2D(
                filters=self.fusion_hidden_dim,
                kernel_size=3,
                data_format=self.data_format,
                use_bias=False,
                dtype=self.dtype_policy,
                name=f"convs_{i}",
            )
            for i in range(len(self.neck_hidden_dims))
        ]
        self.fusion_stage = [
            DepthAnythingFeatureFusionLayer(
                hidden_dim=self.fusion_hidden_dim,
                size=size,
                data_format=self.data_format,
                dtype=self.dtype_policy,
                name=f"fusion_stage_{i}",
            )
            for i, size in enumerate(fusion_sizes)
        ]

    def build(self, inputs_shape):
        outputs_shape = []
        # Reassemble stage.
        for i, shape in enumerate(inputs_shape):
            self.token_to_images[i].build(shape)
            shape = self.token_to_images[i].compute_output_shape(shape)
            self.reassemble_stage[i].build(shape)
            shape = self.reassemble_stage[i].compute_output_shape(shape)
            outputs_shape.append(shape)
        # Convs.
        for i, shape in enumerate(outputs_shape):
            self.convs[i].build(shape)
            shape = self.convs[i].compute_output_shape(shape)
            outputs_shape[i] = shape
        # Fusion stage.
        for i, shape in enumerate(reversed(outputs_shape)):
            self.fusion_stage[i].build(shape)

    def call(self, inputs, training=None):
        # Reassemble stage.
        xs = [
            self.reassemble_stage[i](
                self.token_to_images[i](x), training=training
            )
            for i, x in enumerate(inputs)
        ]
        # Convs.
        xs = [
            self.convs[i](self.paddings[i](x), training=training)
            for i, x in enumerate(xs)
        ]
        # Fusion stage.
        fused_xs = []
        fused_x = None
        for i, x in enumerate(reversed(xs)):
            if fused_x is None:
                fused_x = self.fusion_stage[i](
                    x, residual=None, training=training
                )
            else:
                fused_x = self.fusion_stage[i](
                    fused_x, residual=x, training=training
                )
            fused_xs.append(fused_x)
        return fused_xs

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "patch_size": self.patch_size,
                "image_size": self.image_size,
                "backbone_hidden_dim": self.backbone_hidden_dim,
                "neck_hidden_dims": self.neck_hidden_dims,
                "reassemble_factors": self.reassemble_factors,
                "fusion_hidden_dim": self.fusion_hidden_dim,
                "num_cls_tokens": self.num_cls_tokens,
                "num_register_tokens": self.num_register_tokens,
            }
        )
        return config


class DepthAnythingDepthEstimationHead(layers.Layer):
    """A DepthAnything neck layer.

    Args:
        patch_size: int. The size of one side of each patch.
        patch_height: int. The height of each patch.
        patch_width: int. The width of each patch.
        hidden_dim: int. The number of units in the hidden layers.
        fusion_hidden_dim: int. The number of units in the fusion layers.
        head_hidden_dim: int. The number of units in the head layers.
        head_in_index: int. The index of the feature map to be used as input
            to the head.
        data_format: `None` or str. If specified, either `"channels_last"` or
            `"channels_first"`. The ordering of the dimensions in the
            inputs. `"channels_last"` corresponds to inputs with shape
            `(batch_size, height, width, channels)`
            while `"channels_first"` corresponds to inputs with shape
            `(batch_size, channels, height, width)`. It defaults to the
            `image_data_format` value found in your Keras config file at
            `~/.keras/keras.json`. If you never set it, then it will be
            `"channels_last"`.
        **kwargs: other keyword arguments passed to `keras.layers.Layer`,
            including `name`, `dtype` etc.
    """

    def __init__(
        self,
        patch_size,
        patch_height,
        patch_width,
        fusion_hidden_dim,
        head_hidden_dim,
        head_in_index,
        data_format=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.patch_size = int(patch_size)
        self.patch_height = int(patch_height)
        self.patch_width = int(patch_width)
        self.fusion_hidden_dim = int(fusion_hidden_dim)
        self.head_hidden_dim = int(head_hidden_dim)
        self.head_in_index = int(head_in_index)
        self.data_format = standardize_data_format(data_format)

        # Calculate the interpolate size.
        self.interpolate_size = (
            int(self.patch_height * self.patch_size),
            int(self.patch_width * self.patch_size),
        )

        self.padding1 = layers.ZeroPadding2D(
            padding=(1, 1),
            data_format=self.data_format,
            dtype=self.dtype_policy,
            name="padding1",
        )
        self.conv1 = layers.Conv2D(
            filters=self.fusion_hidden_dim // 2,
            kernel_size=3,
            data_format=self.data_format,
            use_bias=True,
            dtype=self.dtype_policy,
            name="conv1",
        )
        self.padding2 = layers.ZeroPadding2D(
            padding=(1, 1),
            data_format=self.data_format,
            dtype=self.dtype_policy,
            name="padding2",
        )
        self.conv2 = layers.Conv2D(
            filters=self.head_hidden_dim,
            kernel_size=3,
            data_format=self.data_format,
            use_bias=True,
            dtype=self.dtype_policy,
            name="conv2",
        )
        self.activation1 = layers.ReLU(
            dtype=self.dtype_policy, name="activation1"
        )
        self.conv3 = layers.Conv2D(
            filters=1,
            kernel_size=1,
            data_format=self.data_format,
            use_bias=True,
            dtype=self.dtype_policy,
            name="conv3",
        )

    def build(self, inputs_shape):
        inputs_shape = inputs_shape[self.head_in_index]
        self.padding1.build(inputs_shape)
        inputs_shape = self.padding1.compute_output_shape(inputs_shape)
        self.conv1.build(inputs_shape)
        inputs_shape = self.conv1.compute_output_shape(inputs_shape)
        inputs_shape = list(inputs_shape)
        if self.data_format == "channels_last":
            inputs_shape[1] = self.interpolate_size[0]
            inputs_shape[2] = self.interpolate_size[1]
        else:
            inputs_shape[2] = self.interpolate_size[0]
            inputs_shape[3] = self.interpolate_size[1]
        self.padding2.build(inputs_shape)
        inputs_shape = self.padding2.compute_output_shape(inputs_shape)
        self.conv2.build(inputs_shape)
        inputs_shape = self.conv2.compute_output_shape(inputs_shape)
        self.activation1.build(inputs_shape)
        self.conv3.build(inputs_shape)
        inputs_shape = self.conv3.compute_output_shape(inputs_shape)

    def call(self, inputs, training=None):
        x = inputs[self.head_in_index]
        x = self.padding1(x, training=training)
        x = self.conv1(x, training=training)
        x = interpolate(
            x, size=self.interpolate_size, data_format=self.data_format
        )
        x = self.padding2(x, training=training)
        x = self.conv2(x, training=training)
        x = self.activation1(x, training=training)
        return self.conv3(x, training=training)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "patch_size": self.patch_size,
                "patch_height": self.patch_height,
                "patch_width": self.patch_width,
                "fusion_hidden_dim": self.fusion_hidden_dim,
                "head_hidden_dim": self.head_hidden_dim,
                "head_in_index": self.head_in_index,
            }
        )
        return config

    def compute_output_shape(self, input_shape):
        input_shape = input_shape[self.head_in_index]
        if self.data_format == "channels_last":
            output_shape = [
                input_shape[0],
                int(self.patch_height * self.patch_size),
                int(self.patch_width * self.patch_size),
                1,
            ]
        else:
            output_shape = [
                input_shape[0],
                1,
                int(self.patch_height * self.patch_size),
                int(self.patch_width * self.patch_size),
            ]
        return output_shape



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.layers.preprocessing.image_converter import ImageConverter
from keras_hub.src.models.hgnetv2.hgnetv2_backbone import HGNetV2Backbone


@keras_hub_export("keras_hub.layers.HGNetV2ImageConverter")
class HGNetV2ImageConverter(ImageConverter):
    backbone_cls = HGNetV2Backbone



# Metadata for loading pretrained model weights.
backbone_presets = {
    "moonshine_tiny_en": {
        "metadata": {
            "description": (
                "Moonshine tiny model for English speech recognition. "
                "Developed by Useful Sensors for real-time transcription."
            ),
            "params": 27092736,
            "path": "moonshine",
        },
        "kaggle_handle": "kaggle://keras/moonshine/Keras/moonshine_tiny_en/1",
    },
    "moonshine_base_en": {
        "metadata": {
            "description": (
                "Moonshine base model for English speech recognition. "
                "Developed by Useful Sensors for real-time transcription."
            ),
            "params": 61513920,
            "path": "moonshine",
        },
        "kaggle_handle": "kaggle://keras/moonshine/Keras/moonshine_base_en/1",
    },
}



import keras
from keras.src import saving

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.backbone import Backbone
from keras_hub.src.models.mobilenet.mobilenet_backbone import SqueezeAndExcite2D
from keras_hub.src.models.mobilenetv5.mobilenetv5_blocks import (
    MobileNetV5MultiScaleFusionAdapter,
)
from keras_hub.src.models.mobilenetv5.mobilenetv5_builder import (
    MobileNetV5Builder,
)
from keras_hub.src.models.mobilenetv5.mobilenetv5_layers import ConvNormAct
from keras_hub.src.models.mobilenetv5.mobilenetv5_utils import (
    feature_take_indices,
)
from keras_hub.src.models.mobilenetv5.mobilenetv5_utils import round_channels
from keras_hub.src.utils.keras_utils import standardize_data_format


@keras_hub_export("keras_hub.models.MobileNetV5Backbone")
class MobileNetV5Backbone(Backbone):
    """MobileNetV5 backbone network.

    This class represents the backbone of the MobileNetV5 architecture, which
    can be used as a feature extractor for various downstream tasks.

    Args:
        stackwise_block_types: list of list of strings. The block type for each
            block in each stack.
        stackwise_num_blocks: list of ints. The number of blocks for each
            stack.
        stackwise_num_filters: list of list of ints. The number of filters for
            each block in each stack.
        stackwise_strides: list of list of ints. The stride for each block in
            each stack.
        stackwise_act_layers: list of list of strings. The activation function
            for each block in each stack.
        stackwise_exp_ratios: list of list of floats. The expansion ratio for
            each block in each stack.
        stackwise_se_ratios: list of list of floats. The SE ratio for each
            block in each stack.
        stackwise_dw_kernel_sizes: list of list of ints. The depthwise kernel
            size for each block in each stack.
        stackwise_dw_start_kernel_sizes: list of list of ints. The start
            depthwise kernel size for each `uir` block in each stack.
        stackwise_dw_end_kernel_sizes: list of list of ints. The end depthwise
            kernel size for each `uir` block in each stack.
        stackwise_exp_kernel_sizes: list of list of ints. The expansion kernel
            size for each `er` block in each stack.
        stackwise_pw_kernel_sizes: list of list of ints. The pointwise kernel
            size for each `er` block in each stack.
        stackwise_num_heads: list of list of ints. The number of heads for each
            `mqa` or `mha` block in each stack.
        stackwise_key_dims: list of list of ints. The key dimension for each
            `mqa` or `mha` block in each stack.
        stackwise_value_dims: list of list of ints. The value dimension for each
            `mqa` or `mha` block in each stack.
        stackwise_kv_strides: list of list of ints. The key-value stride for
            each `mqa` or `mha` block in each stack.
        stackwise_use_cpe: list of list of bools. Whether to use conditional
            position encoding for each `mqa` or `mha` block in each stack.
        filters: int. The number of input channels.
        stem_size: int. The number of channels in the stem convolution.
        stem_bias: bool. If `True`, a bias term is used in the stem
            convolution.
        fix_stem: bool. If `True`, the stem size is not rounded.
        num_features: int. The number of output features, used when `use_msfa`
            is `True`.
        pad_type: str. The padding type for convolutions.
        use_msfa: bool. If `True`, the Multi-Scale Fusion Adapter is used.
        msfa_indices: tuple. The indices of the feature maps to be used by the
            MSFA.
        msfa_output_resolution: int. The output resolution of the MSFA.
        act_layer: str. The activation function to use.
        norm_layer: str. The normalization layer to use.
        se_layer: keras.layers.Layer. The Squeeze-and-Excitation layer to use.
        se_from_exp: bool. If `True`, SE channel reduction is based on the
            expanded channels.
        round_chs_fn: callable. A function to round the number of channels.
        drop_path_rate: float. The stochastic depth rate.
        layer_scale_init_value: float. The initial value for layer scale.
        image_shape: tuple. The shape of the input image. Defaults to
            `(None, None, 3)`.
        data_format: str, The data format of the image channels. Can be either
            `"channels_first"` or `"channels_last"`. If `None` is specified,
            it will use the `image_data_format` value found in your Keras
            config file at `~/.keras/keras.json`. Defaults to `None`.
        dtype: `None` or str or `keras.mixed_precision.DTypePolicy`. The dtype
            to use for the model's computations and weights. Defaults to `None`.

    Example:
    ```python
    import keras
    from keras_hub.models import MobileNetV5Backbone

    # Randomly initialized backbone with a custom config.
    model_args = {
        "stackwise_block_types": [["er"], ["uir", "uir"]],
        "stackwise_num_blocks": [1, 2],
        "stackwise_num_filters": [[24], [48, 48]],
        "stackwise_strides": [[2], [2, 1]],
        "stackwise_act_layers": [["relu"], ["relu", "relu"]],
        "stackwise_exp_ratios": [[4.0], [6.0, 6.0]],
        "stackwise_se_ratios": [[0.0], [0.0, 0.0]],
        "stackwise_dw_kernel_sizes": [[0], [5, 5]],
        "stackwise_dw_start_kernel_sizes": [[0], [0, 0]],
        "stackwise_dw_end_kernel_sizes": [[0], [0, 0]],
        "stackwise_exp_kernel_sizes": [[3], [0, 0]],
        "stackwise_pw_kernel_sizes": [[1], [0, 0]],
        "stackwise_num_heads": [[0], [0, 0]],
        "stackwise_key_dims": [[0], [0, 0]],
        "stackwise_value_dims": [[0], [0, 0]],
        "stackwise_kv_strides": [[0], [0, 0]],
        "stackwise_use_cpe": [[False], [False, False]],
        "use_msfa": False,
    }
    model = MobileNetV5Backbone(**model_args)
    input_data = keras.ops.ones((1, 224, 224, 3))
    output = model(input_data)

    # Load the backbone from a preset and run a prediction.
    backbone = MobileNetV5Backbone.from_preset("mobilenetv5_300m_gemma3n")

    # Expected output shape = (1, 16, 16, 2048).
    outputs = backbone.predict(keras.ops.ones((1, 224, 224, 3)))
    ```
    """

    def __init__(
        self,
        stackwise_block_types,
        stackwise_num_blocks,
        stackwise_num_filters,
        stackwise_strides,
        stackwise_act_layers,
        stackwise_exp_ratios,
        stackwise_se_ratios,
        stackwise_dw_kernel_sizes,
        stackwise_dw_start_kernel_sizes,
        stackwise_dw_end_kernel_sizes,
        stackwise_exp_kernel_sizes,
        stackwise_pw_kernel_sizes,
        stackwise_num_heads,
        stackwise_key_dims,
        stackwise_value_dims,
        stackwise_kv_strides,
        stackwise_use_cpe,
        filters=3,
        stem_size=16,
        stem_bias=True,
        fix_stem=False,
        num_features=2048,
        pad_type="same",
        use_msfa=True,
        msfa_indices=(-2, -1),
        msfa_output_resolution=16,
        act_layer="gelu",
        norm_layer="rms_norm",
        se_layer=SqueezeAndExcite2D,
        se_from_exp=True,
        round_chs_fn=round_channels,
        drop_path_rate=0.0,
        layer_scale_init_value=None,
        image_shape=(None, None, 3),
        data_format=None,
        dtype=None,
        **kwargs,
    ):
        data_format = standardize_data_format(data_format)
        channel_axis = -1 if data_format == "channels_last" else 1
        block_args = []
        for i in range(len(stackwise_num_blocks)):
            stack_args = []
            for j in range(stackwise_num_blocks[i]):
                block_type = stackwise_block_types[i][j]
                args = {
                    "block_type": block_type,
                    "out_chs": stackwise_num_filters[i][j],
                    "stride": stackwise_strides[i][j],
                    "act_layer": stackwise_act_layers[i][j],
                }
                if block_type == "ir":
                    args.update(
                        {
                            "dw_kernel_size": stackwise_dw_kernel_sizes[i][j],
                            "exp_ratio": stackwise_exp_ratios[i][j],
                            "se_ratio": stackwise_se_ratios[i][j],
                        }
                    )
                elif block_type == "uir":
                    args.update(
                        {
                            "dw_kernel_size_mid": stackwise_dw_kernel_sizes[i][
                                j
                            ],
                            "dw_kernel_size_start": stackwise_dw_start_kernel_sizes[  # noqa: E501
                                i
                            ][j],
                            "dw_kernel_size_end": stackwise_dw_end_kernel_sizes[
                                i
                            ][j],
                            "exp_ratio": stackwise_exp_ratios[i][j],
                            "se_ratio": stackwise_se_ratios[i][j],
                        }
                    )
                elif block_type == "er":
                    args.update(
                        {
                            "exp_kernel_size": stackwise_exp_kernel_sizes[i][j],
                            "pw_kernel_size": stackwise_pw_kernel_sizes[i][j],
                            "exp_ratio": stackwise_exp_ratios[i][j],
                            "se_ratio": stackwise_se_ratios[i][j],
                        }
                    )
                elif block_type in ("mqa", "mha"):
                    args.update(
                        {
                            "num_heads": stackwise_num_heads[i][j],
                            "key_dim": stackwise_key_dims[i][j],
                            "value_dim": stackwise_value_dims[i][j],
                            "kv_stride": stackwise_kv_strides[i][j],
                            "use_cpe": stackwise_use_cpe[i][j],
                        }
                    )
                stack_args.append(args)
            block_args.append(stack_args)

        # === Layers ===
        if not fix_stem:
            stem_size = round_chs_fn(stem_size)
        conv_stem = ConvNormAct(
            stem_size,
            kernel_size=3,
            stride=2,
            pad_type=pad_type,
            bias=stem_bias,
            norm_layer=norm_layer,
            act_layer=act_layer,
            name="conv_stem",
            data_format=data_format,
            channel_axis=channel_axis,
            dtype=dtype,
        )
        builder = MobileNetV5Builder(
            output_stride=32,
            pad_type=pad_type,
            round_chs_fn=round_chs_fn,
            se_from_exp=se_from_exp,
            act_layer=act_layer,
            norm_layer=norm_layer,
            se_layer=se_layer,
            drop_path_rate=drop_path_rate,
            layer_scale_init_value=layer_scale_init_value,
            data_format=data_format,
            channel_axis=channel_axis,
            dtype=dtype,
        )
        blocks = builder(stem_size, block_args)
        feature_info = builder.features
        msfa = None
        if use_msfa:
            msfa_indices_calc, _ = feature_take_indices(
                len(feature_info), msfa_indices
            )
            msfa_in_chs = [
                feature_info[mi]["num_chs"] for mi in msfa_indices_calc
            ]
            msfa = MobileNetV5MultiScaleFusionAdapter(
                in_chs=msfa_in_chs,
                filters=num_features,
                output_resolution=msfa_output_resolution,
                norm_layer=norm_layer,
                act_layer=act_layer,
                name="msfa",
                channel_axis=channel_axis,
                data_format=data_format,
                dtype=dtype,
            )

        # === Functional Model ===
        image_input = keras.layers.Input(shape=image_shape)
        x = conv_stem(image_input)
        if use_msfa:
            intermediates = []
            feat_idx = 0
            if feat_idx in msfa_indices_calc:
                intermediates.append(x)

            for stage in blocks:
                for block in stage:
                    x = block(x)
                feat_idx += 1
                if feat_idx in msfa_indices_calc:
                    intermediates.append(x)
            x = msfa(intermediates)
        else:
            for stage in blocks:
                for block in stage:
                    x = block(x)

        super().__init__(inputs=image_input, outputs=x, dtype=dtype, **kwargs)

        # === Config ===
        self.stackwise_block_types = stackwise_block_types
        self.stackwise_num_blocks = stackwise_num_blocks
        self.stackwise_num_filters = stackwise_num_filters
        self.stackwise_strides = stackwise_strides
        self.stackwise_act_layers = stackwise_act_layers
        self.stackwise_exp_ratios = stackwise_exp_ratios
        self.stackwise_se_ratios = stackwise_se_ratios
        self.stackwise_dw_kernel_sizes = stackwise_dw_kernel_sizes
        self.stackwise_dw_start_kernel_sizes = stackwise_dw_start_kernel_sizes
        self.stackwise_dw_end_kernel_sizes = stackwise_dw_end_kernel_sizes
        self.stackwise_exp_kernel_sizes = stackwise_exp_kernel_sizes
        self.stackwise_pw_kernel_sizes = stackwise_pw_kernel_sizes
        self.stackwise_num_heads = stackwise_num_heads
        self.stackwise_key_dims = stackwise_key_dims
        self.stackwise_value_dims = stackwise_value_dims
        self.stackwise_kv_strides = stackwise_kv_strides
        self.stackwise_use_cpe = stackwise_use_cpe
        self.filters = filters
        self.stem_size = stem_size
        self.stem_bias = stem_bias
        self.fix_stem = fix_stem
        self.num_features = num_features
        self.pad_type = pad_type
        self.use_msfa = use_msfa
        self.msfa_indices = msfa_indices
        self.msfa_output_resolution = msfa_output_resolution
        self.act_layer = act_layer
        self.norm_layer = norm_layer
        self.se_layer = se_layer
        self.se_from_exp = se_from_exp
        self.round_chs_fn = round_chs_fn
        self.drop_path_rate = drop_path_rate
        self.layer_scale_init_value = layer_scale_init_value
        self.image_shape = image_shape
        self.data_format = data_format
        self.channel_axis = channel_axis

    def get_config(self):
        config = {
            "stackwise_block_types": self.stackwise_block_types,
            "stackwise_num_blocks": self.stackwise_num_blocks,
            "stackwise_num_filters": self.stackwise_num_filters,
            "stackwise_strides": self.stackwise_strides,
            "stackwise_act_layers": self.stackwise_act_layers,
            "stackwise_exp_ratios": self.stackwise_exp_ratios,
            "stackwise_se_ratios": self.stackwise_se_ratios,
            "stackwise_dw_kernel_sizes": self.stackwise_dw_kernel_sizes,
            "stackwise_dw_start_kernel_sizes": self.stackwise_dw_start_kernel_sizes,  # noqa: E501
            "stackwise_dw_end_kernel_sizes": self.stackwise_dw_end_kernel_sizes,
            "stackwise_exp_kernel_sizes": self.stackwise_exp_kernel_sizes,
            "stackwise_pw_kernel_sizes": self.stackwise_pw_kernel_sizes,
            "stackwise_num_heads": self.stackwise_num_heads,
            "stackwise_key_dims": self.stackwise_key_dims,
            "stackwise_value_dims": self.stackwise_value_dims,
            "stackwise_kv_strides": self.stackwise_kv_strides,
            "stackwise_use_cpe": self.stackwise_use_cpe,
            "filters": self.filters,
            "stem_size": self.stem_size,
            "stem_bias": self.stem_bias,
            "fix_stem": self.fix_stem,
            "num_features": self.num_features,
            "pad_type": self.pad_type,
            "use_msfa": self.use_msfa,
            "msfa_indices": self.msfa_indices,
            "msfa_output_resolution": self.msfa_output_resolution,
            "act_layer": self.act_layer,
            "norm_layer": self.norm_layer,
            "se_from_exp": self.se_from_exp,
            "drop_path_rate": self.drop_path_rate,
            "layer_scale_init_value": self.layer_scale_init_value,
            "image_shape": self.image_shape,
            "data_format": self.data_format,
        }
        if self.round_chs_fn is not round_channels:
            config["round_chs_fn"] = saving.serialize_keras_object(
                self.round_chs_fn
            )
        if self.se_layer is not SqueezeAndExcite2D:
            config["se_layer"] = saving.serialize_keras_object(self.se_layer)
        return config

    @classmethod
    def from_config(cls, config):
        if "round_chs_fn" in config:
            config["round_chs_fn"] = saving.deserialize_keras_object(
                config["round_chs_fn"]
            )
        if "se_layer" in config:
            config["se_layer"] = saving.deserialize_keras_object(
                config["se_layer"]
            )
        return cls(**config)



from keras_hub.src.models.falcon.falcon_backbone import FalconBackbone
from keras_hub.src.models.falcon.falcon_presets import backbone_presets
from keras_hub.src.utils.preset_utils import register_presets

register_presets(backbone_presets, FalconBackbone)



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.cspnet.cspnet_backbone import CSPNetBackbone
from keras_hub.src.models.cspnet.cspnet_image_converter import (
    CSPNetImageConverter,
)
from keras_hub.src.models.image_classifier_preprocessor import (
    ImageClassifierPreprocessor,
)


@keras_hub_export("keras_hub.models.CSPNetImageClassifierPreprocessor")
class CSPNetImageClassifierPreprocessor(ImageClassifierPreprocessor):
    backbone_cls = CSPNetBackbone
    image_converter_cls = CSPNetImageConverter



import numpy as np

from keras_hub.src.models.albert.albert_backbone import AlbertBackbone
from keras_hub.src.utils.preset_utils import get_file

backbone_cls = AlbertBackbone


def convert_backbone_config(transformers_config):
    return {
        "vocabulary_size": transformers_config["vocab_size"],
        "num_layers": transformers_config["num_hidden_layers"],
        "num_heads": transformers_config["num_attention_heads"],
        "embedding_dim": transformers_config["embedding_size"],
        "hidden_dim": transformers_config["hidden_size"],
        "intermediate_dim": transformers_config["intermediate_size"],
        "num_groups": transformers_config["num_hidden_groups"],
        "num_inner_repetitions": transformers_config["inner_group_num"],
        "dropout": transformers_config["attention_probs_dropout_prob"],
        "max_sequence_length": transformers_config["max_position_embeddings"],
        "num_segments": transformers_config["type_vocab_size"],
    }


def convert_weights(backbone, loader, transformers_config):
    # Embeddings
    loader.port_weight(
        keras_variable=backbone.token_embedding.embeddings,
        hf_weight_key="albert.embeddings.word_embeddings.weight",
    )
    loader.port_weight(
        keras_variable=backbone.position_embedding.position_embeddings,
        hf_weight_key="albert.embeddings.position_embeddings.weight",
    )
    loader.port_weight(
        keras_variable=backbone.segment_embedding.embeddings,
        hf_weight_key="albert.embeddings.token_type_embeddings.weight",
    )

    # Normalization
    loader.port_weight(
        keras_variable=backbone.embeddings_layer_norm.gamma,
        hf_weight_key="albert.embeddings.LayerNorm.weight",
    )
    loader.port_weight(
        keras_variable=backbone.embeddings_layer_norm.beta,
        hf_weight_key="albert.embeddings.LayerNorm.bias",
    )

    # Encoder Embeddings
    loader.port_weight(
        keras_variable=backbone.embeddings_projection.kernel,
        hf_weight_key="albert.encoder.embedding_hidden_mapping_in.weight",
        hook_fn=lambda hf_tensor, _: np.transpose(hf_tensor, axes=(1, 0)),
    )
    loader.port_weight(
        keras_variable=backbone.embeddings_projection.bias,
        hf_weight_key="albert.encoder.embedding_hidden_mapping_in.bias",
    )

    # Encoder Group Layers
    for group_idx in range(backbone.num_groups):
        for inner_layer_idx in range(backbone.num_inner_repetitions):
            keras_group = backbone.get_layer(
                f"group_{group_idx}_inner_layer_{inner_layer_idx}"
            )
            hf_group_prefix = (
                "albert.encoder.albert_layer_groups."
                f"{group_idx}.albert_layers.{inner_layer_idx}."
            )

            loader.port_weight(
                keras_variable=keras_group._self_attention_layer.query_dense.kernel,
                hf_weight_key=f"{hf_group_prefix}attention.query.weight",
                hook_fn=lambda hf_tensor, keras_shape: np.reshape(
                    np.transpose(hf_tensor), keras_shape
                ),
            )
            loader.port_weight(
                keras_variable=keras_group._self_attention_layer.query_dense.bias,
                hf_weight_key=f"{hf_group_prefix}attention.query.bias",
                hook_fn=lambda hf_tensor, keras_shape: np.reshape(
                    hf_tensor, keras_shape
                ),
            )
            loader.port_weight(
                keras_variable=keras_group._self_attention_layer.key_dense.kernel,
                hf_weight_key=f"{hf_group_prefix}attention.key.weight",
                hook_fn=lambda hf_tensor, keras_shape: np.reshape(
                    np.transpose(hf_tensor), keras_shape
                ),
            )
            loader.port_weight(
                keras_variable=keras_group._self_attention_layer.key_dense.bias,
                hf_weight_key=f"{hf_group_prefix}attention.key.bias",
                hook_fn=lambda hf_tensor, keras_shape: np.reshape(
                    hf_tensor, keras_shape
                ),
            )
            loader.port_weight(
                keras_variable=keras_group._self_attention_layer.value_dense.kernel,
                hf_weight_key=f"{hf_group_prefix}attention.value.weight",
                hook_fn=lambda hf_tensor, keras_shape: np.reshape(
                    np.transpose(hf_tensor), keras_shape
                ),
            )
            loader.port_weight(
                keras_variable=keras_group._self_attention_layer.value_dense.bias,
                hf_weight_key=f"{hf_group_prefix}attention.value.bias",
                hook_fn=lambda hf_tensor, keras_shape: np.reshape(
                    hf_tensor, keras_shape
                ),
            )
            loader.port_weight(
                keras_variable=keras_group._self_attention_layer.output_dense.kernel,
                hf_weight_key=f"{hf_group_prefix}attention.dense.weight",
                hook_fn=lambda hf_tensor, keras_shape: np.reshape(
                    np.transpose(hf_tensor), keras_shape
                ),
            )
            loader.port_weight(
                keras_variable=keras_group._self_attention_layer.output_dense.bias,
                hf_weight_key=f"{hf_group_prefix}attention.dense.bias",
                hook_fn=lambda hf_tensor, keras_shape: np.reshape(
                    hf_tensor, keras_shape
                ),
            )
            loader.port_weight(
                keras_variable=keras_group._self_attention_layer_norm.gamma,
                hf_weight_key=f"{hf_group_prefix}attention.LayerNorm.weight",
            )
            loader.port_weight(
                keras_variable=keras_group._self_attention_layer_norm.beta,
                hf_weight_key=f"{hf_group_prefix}attention.LayerNorm.bias",
            )
            loader.port_weight(
                keras_variable=keras_group._feedforward_intermediate_dense.kernel,
                hf_weight_key=f"{hf_group_prefix}ffn.weight",
                hook_fn=lambda hf_tensor, _: np.transpose(
                    hf_tensor, axes=(1, 0)
                ),
            )
            loader.port_weight(
                keras_variable=keras_group._feedforward_intermediate_dense.bias,
                hf_weight_key=f"{hf_group_prefix}ffn.bias",
            )
            loader.port_weight(
                keras_variable=keras_group._feedforward_output_dense.kernel,
                hf_weight_key=f"{hf_group_prefix}ffn_output.weight",
                hook_fn=lambda hf_tensor, _: np.transpose(
                    hf_tensor, axes=(1, 0)
                ),
            )
            loader.port_weight(
                keras_variable=keras_group._feedforward_output_dense.bias,
                hf_weight_key=f"{hf_group_prefix}ffn_output.bias",
            )
            loader.port_weight(
                keras_variable=keras_group._feedforward_layer_norm.gamma,
                hf_weight_key=f"{hf_group_prefix}full_layer_layer_norm.weight",
            )
            loader.port_weight(
                keras_variable=keras_group._feedforward_layer_norm.beta,
                hf_weight_key=f"{hf_group_prefix}full_layer_layer_norm.bias",
            )

    # Pooler
    loader.port_weight(
        keras_variable=backbone.pooled_dense.kernel,
        hf_weight_key="albert.pooler.weight",
        hook_fn=lambda hf_tensor, _: np.transpose(hf_tensor, axes=(1, 0)),
    )
    loader.port_weight(
        keras_variable=backbone.pooled_dense.bias,
        hf_weight_key="albert.pooler.bias",
    )


def convert_tokenizer(cls, preset, **kwargs):
    return cls(get_file(preset, "spiece.model"), **kwargs)



import keras

from keras_hub.src.models.mobilenet.util import adjust_channels


def num_groups(group_size, channels):
    if not group_size:
        return 1
    else:
        if channels % group_size != 0:
            raise ValueError(
                f"Number of channels ({channels}) must be divisible by "
                "group size ({group_size})."
            )
        return channels // group_size


def parse_ksize(ss):
    if ss.isdigit():
        return int(ss)
    else:
        return [int(k) for k in ss.split(".")]


def round_channels(
    channels, multiplier=1.0, divisor=8, channel_min=None, round_limit=0.9
):
    if not multiplier:
        return channels
    return adjust_channels(channels * multiplier, divisor, channel_min)


def feature_take_indices(num_stages, indices):
    if not isinstance(indices, (tuple, list)):
        indices = (indices,)
    if any(i < 0 for i in indices):
        indices = [i if i >= 0 else num_stages + i for i in indices]
    return indices, max(indices)


class SelectAdaptivePool2d(keras.layers.Layer):
    """A layer that selects and applies a 2D adaptive pooling strategy.

    This layer supports various pooling types like average, max, or a
    combination of both. It can also flatten the output.

    Args:
        pool_type: str. The type of pooling to apply. One of `"avg"`, `"max"`,
            `"avgmax"`, `"catavgmax"`, or `""` (identity).
        flatten: bool. If `True`, the output is flattened after pooling.
        data_format: str. The format of the input data, either
            `"channels_last"` or `"channels_first"`.
        channel_axis: int. The axis representing the channels in the input
            tensor.
    """

    def __init__(
        self,
        pool_type="avg",
        flatten=False,
        data_format=None,
        channel_axis=None,
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.pool_type = pool_type.lower()
        self.flatten = flatten
        self.data_format = data_format
        self.channels_axis = channel_axis
        self.pool = None
        self.pool_avg = None
        self.pool_max = None
        self.pool_cat = None
        self.flatten_layer = None
        if self.pool_type not in ("avg", "max", "avgmax", "catavgmax", ""):
            raise ValueError(f"Invalid pool type: {self.pool_type}")

    def build(self, input_shape):
        if self.pool_type == "avg":
            self.pool = keras.layers.GlobalAveragePooling2D(
                data_format=self.data_format,
                keepdims=not self.flatten,
                dtype=self.dtype_policy,
            )
        elif self.pool_type == "max":
            self.pool = keras.layers.GlobalMaxPooling2D(
                data_format=self.data_format,
                keepdims=not self.flatten,
                dtype=self.dtype_policy,
            )
        elif self.pool_type in ("avgmax", "catavgmax"):
            self.pool_avg = keras.layers.GlobalAveragePooling2D(
                data_format=self.data_format,
                keepdims=not self.flatten,
                dtype=self.dtype_policy,
            )
            self.pool_max = keras.layers.GlobalMaxPooling2D(
                data_format=self.data_format,
                keepdims=not self.flatten,
                dtype=self.dtype_policy,
            )
            if self.pool_type == "catavgmax":
                axis = 1 if self.data_format == "channels_first" else -1
                self.pool_cat = keras.layers.Concatenate(
                    axis=axis, dtype=self.dtype_policy
                )
        elif not self.pool_type:
            self.pool = keras.layers.Identity(dtype=self.dtype_policy)
            if self.flatten:
                self.flatten_layer = keras.layers.Flatten(
                    dtype=self.dtype_policy
                )
        super().build(input_shape)

    def call(self, x):
        if self.pool_type in ("avg", "max"):
            return self.pool(x)
        elif self.pool_type == "avgmax":
            x_avg = self.pool_avg(x)
            x_max = self.pool_max(x)
            return 0.5 * (x_avg + x_max)
        elif self.pool_type == "catavgmax":
            x_avg = self.pool_avg(x)
            x_max = self.pool_max(x)
            return self.pool_cat([x_avg, x_max])
        elif not self.pool_type:
            x = self.pool(x)
            if self.flatten_layer:
                x = self.flatten_layer(x)
            return x
        return x

    def feat_mult(self):
        return 2 if self.pool_type == "catavgmax" else 1

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "pool_type": self.pool_type,
                "flatten": self.flatten,
                "data_format": self.data_format,
            }
        )
        return config



from keras_hub.src.models.flux.flux_model import FluxBackbone
from keras_hub.src.models.flux.flux_presets import presets
from keras_hub.src.utils.preset_utils import register_presets

register_presets(presets, FluxBackbone)



import keras
from keras import ops
from packaging import version

from keras_hub.src.layers.modeling.rotary_embedding import RotaryEmbedding
from keras_hub.src.models.roformer_v2.roformer_v2_attention import (
    RoformerAttention,
)


class ESMRotaryEmbedding(RotaryEmbedding):
    def _compute_cos_sin_embedding(self, x, position=1):
        dim = x.shape[-1]
        inv_freq = self.scaling_factor / (
            self.max_wavelength ** (ops.arange(0, dim, 2, dtype=x.dtype) / dim)
        )
        t = ops.arange(x.shape[position], dtype=x.dtype)
        freqs = ops.outer(t, inv_freq)
        emb = ops.concatenate((freqs, freqs), axis=-1)

        cos_emb = ops.cos(emb)[None, :, None, :]
        sin_emb = ops.sin(emb)[None, :, None, :]
        return cos_emb, sin_emb

    def call(self, q, k, position=1):
        cos_emb, sin_emb = self._compute_cos_sin_embedding(q, position)

        return (
            self.apply_rotary_pos_emb(q, cos_emb, sin_emb),
            self.apply_rotary_pos_emb(k, cos_emb, sin_emb),
        )

    def rotate_half(self, x):
        x1, x2 = ops.split(x, 2, -1)
        return ops.concatenate((-x2, x1), axis=-1)

    def apply_rotary_pos_emb(self, x, cos, sin):
        cos = cos[:, : x.shape[1], :, :]
        sin = sin[:, : x.shape[1], :, :]

        return (x * cos) + (self.rotate_half(x) * sin)


class EsmSelfAttention(RoformerAttention):
    """MultiHeadAttention by ESM2

    Referred to the implementation of HuggingFace.
    In fact, this part of the calculation is exactly the same as RoFormer.
    Only the calculation of the rotary part is different.
    """

    def __init__(self, use_rotary=True, **kwargs):
        super().__init__(**kwargs)
        self.use_rotary = use_rotary

    def build(self, input_shape):
        super().build(input_shape)
        if self.use_rotary:
            self.rotary_embedding_layer = ESMRotaryEmbedding(
                max_wavelength=self.max_wavelength, dtype=self.dtype_policy
            )
            self.rotary_embedding_layer.build([])

    def call(self, x, attention_mask=None):
        qw = self.q_dense(x)
        kw = self.k_dense(x)
        vw = self.v_dense(x)

        b, s = ops.shape(qw)[:2]
        qw = ops.reshape(qw, (b, s, self.heads, self.head_size))
        kw = ops.reshape(kw, (b, s, self.heads, self.head_size))
        vw = ops.reshape(vw, (b, s, self.heads, self.head_size))

        if self.use_rotary:
            qw, kw = self.rotary_embedding_layer(qw, kw)
        if version.parse(keras.__version__) < version.parse("3.6"):
            raise ValueError("Please make sure your Keras version is >=3.6.")
        flash_attention = keras.config.is_flash_attention_enabled()
        attention_mask = ops.reshape(attention_mask, [b, 1, s, 1])
        if keras.config.backend() == "torch":
            attention_mask = ops.repeat(attention_mask, s, -1)
            attention_mask = ops.transpose(attention_mask, [0, 1, 3, 2])
        o = ops.dot_product_attention(
            qw, kw, vw, mask=attention_mask, flash_attention=flash_attention
        )
        return self.o_dense(ops.reshape(o, [b, s, -1]))

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "use_rotary": self.use_rotary,
            }
        )
        return config



from keras import backend
from keras import ops

from keras_hub.src.utils.keras_utils import standardize_data_format


def interpolate(x, size, data_format=None):
    """Performs a backend-agnostic version of Torch's `F.interpolate`.

    Args:
        x: A 4D image tensor.
        size: A tuple of 2 integers, `(height, width)`.
        data_format: One of `channels_last` or `channels_first`.
    """
    data_format = standardize_data_format(data_format)
    if backend.backend() == "jax":
        import jax

        if data_format == "channels_first":
            x = ops.transpose(x, (0, 2, 3, 1))
        scale = ops.convert_to_tensor(
            [
                (size[0] - 1.0) / (x.shape[1] - 1.0),
                (size[1] - 1.0) / (x.shape[2] - 1.0),
            ]
        )
        translation = -(scale / 2.0 - 0.5)
        x = jax.image.scale_and_translate(
            x,
            (x.shape[0], *size, x.shape[-1]),
            method="bilinear",
            scale=scale,
            spatial_dims=(1, 2),
            translation=translation,
            antialias=False,
        )
        if data_format == "channels_first":
            x = ops.transpose(x, (0, 3, 1, 2))
    elif backend.backend() == "tensorflow":
        import tensorflow as tf

        if data_format == "channels_first":
            x = ops.transpose(x, (0, 2, 3, 1))
        x = tf.compat.v1.image.resize(
            x,
            size=size,
            method="bilinear",
            align_corners=True,
        )
        if data_format == "channels_first":
            x = ops.transpose(x, (0, 3, 1, 2))
    elif backend.backend() == "torch":
        import torch.nn.functional as F

        if data_format == "channels_last":
            x = ops.transpose(x, (0, 3, 1, 2))
        x = F.interpolate(x, size=size, mode="bilinear", align_corners=True)
        if data_format == "channels_last":
            x = ops.transpose(x, (0, 2, 3, 1))
    else:
        raise NotImplementedError(f"Unsupported backend: {backend.backend()}")
    return x



"""RoBERTa model preset configurations."""

backbone_presets = {
    "roberta_base_en": {
        "metadata": {
            "description": (
                "12-layer RoBERTa model where case is maintained."
                "Trained on English Wikipedia, BooksCorpus, CommonCraw, and "
                "OpenWebText."
            ),
            "params": 124052736,
            "path": "roberta",
        },
        "kaggle_handle": "kaggle://keras/roberta/keras/roberta_base_en/3",
    },
    "roberta_large_en": {
        "metadata": {
            "description": (
                "24-layer RoBERTa model where case is maintained."
                "Trained on English Wikipedia, BooksCorpus, CommonCraw, and "
                "OpenWebText."
            ),
            "params": 354307072,
            "path": "roberta",
        },
        "kaggle_handle": "kaggle://keras/roberta/keras/roberta_large_en/3",
    },
}



import keras
from keras import ops

from keras_hub.src.layers.modeling.transformer_layer_utils import (
    compute_causal_mask,
)
from keras_hub.src.layers.modeling.transformer_layer_utils import (
    merge_padding_and_attention_mask,
)
from keras_hub.src.models.gemma3.gemma3_attention import CachedGemma3Attention
from keras_hub.src.models.gemma3.rms_normalization import RMSNormalization


class Gemma3DecoderBlock(keras.layers.Layer):
    """Transformer decoder layer for Gemma3.

    This decoder layer is the same as the layer used for Gemma and Gemma2.
    However, there are a few key differences. Firstly, image tokens have
    bidirectional masking. Additionally, this layer exposes the following args:

    `use_query_key_norm`: bool. If True, apply RMS normalization on query
        and key. For Gemma3, this is True.
    `rope_wavelength`: float. Configurable value for RoPE wavelength. Gemma3
        uses 10K for local attention layers and 1M for global attention layers.
    `gate_dim_reduction`: int. In the gating layers, the output dimension is
        `intermediate_dim // gate_dim_reduction`. For Gemma and Gemma2, this
        value is 2. For Gemma3, it is 1.
    """

    def __init__(
        self,
        hidden_dim,
        intermediate_dim,
        head_dim,
        num_query_heads,
        num_key_value_heads,
        query_head_dim_normalize=True,
        use_query_key_norm=False,
        use_post_ffw_norm=False,
        use_post_attention_norm=False,
        gate_dim_reduction=2,
        logit_soft_cap=None,
        use_sliding_window_attention=False,
        sliding_window_size=4096,
        layer_norm_epsilon=1e-6,
        rope_wavelength=10_000.0,
        rope_scaling_factor=1.0,
        use_bidirectional_attention=False,
        dropout=0,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.hidden_dim = hidden_dim
        self.intermediate_dim = intermediate_dim
        self.head_dim = head_dim
        self.num_query_heads = num_query_heads
        self.num_key_value_heads = num_key_value_heads
        self.query_head_dim_normalize = query_head_dim_normalize
        self.use_query_key_norm = use_query_key_norm
        self.use_post_ffw_norm = use_post_ffw_norm
        self.use_post_attention_norm = use_post_attention_norm
        self.gate_dim_reduction = gate_dim_reduction
        self.logit_soft_cap = logit_soft_cap
        self.use_sliding_window_attention = use_sliding_window_attention
        self.sliding_window_size = sliding_window_size
        self.layer_norm_epsilon = layer_norm_epsilon
        self.rope_wavelength = rope_wavelength
        self.rope_scaling_factor = rope_scaling_factor
        self.use_bidirectional_attention = use_bidirectional_attention
        self.dropout = dropout

        self.pre_attention_norm = RMSNormalization(
            epsilon=self.layer_norm_epsilon,
            dtype=self.dtype_policy,
            name="pre_attention_norm",
        )

        if use_post_attention_norm:
            self.post_attention_norm = RMSNormalization(
                epsilon=self.layer_norm_epsilon,
                dtype=self.dtype_policy,
                name="post_attention_norm",
            )

        self.attention = CachedGemma3Attention(
            head_dim=head_dim,
            num_query_heads=num_query_heads,
            num_key_value_heads=num_key_value_heads,
            use_query_key_norm=use_query_key_norm,
            logit_soft_cap=logit_soft_cap,
            use_sliding_window_attention=use_sliding_window_attention,
            sliding_window_size=sliding_window_size,
            query_head_dim_normalize=True,
            rope_wavelength=rope_wavelength,
            rope_scaling_factor=rope_scaling_factor,
            dropout=dropout,
            use_bidirectional_attention=use_bidirectional_attention,
            dtype=self.dtype_policy,
            name="attention",
        )

        if self.dropout > 0:
            self.attention_dropout = keras.layers.Dropout(rate=dropout)
            self.feedforward_dropout = keras.layers.Dropout(rate=dropout)

        self.pre_ffw_norm = RMSNormalization(
            epsilon=self.layer_norm_epsilon,
            dtype=self.dtype_policy,
            name="pre_ffw_norm",
        )

        if use_post_ffw_norm:
            self.post_ffw_norm = RMSNormalization(
                epsilon=self.layer_norm_epsilon,
                dtype=self.dtype_policy,
                name="post_ffw_norm",
            )

        self.gating_ffw = keras.layers.EinsumDense(
            equation="btd,df->btf",
            output_shape=(None, self.intermediate_dim // gate_dim_reduction),
            dtype=self.dtype_policy,
            name="ffw_gating",
        )

        self.gating_ffw_2 = keras.layers.EinsumDense(
            equation="btd,df->btf",
            output_shape=(None, self.intermediate_dim // gate_dim_reduction),
            dtype=self.dtype_policy,
            name="ffw_gating_2",
        )

        self.ffw_linear = keras.layers.EinsumDense(
            equation="btf,fd->btd",
            output_shape=(None, self.hidden_dim),
            dtype=self.dtype_policy,
            name="ffw_linear",
        )

    def build(self, input_shape):
        self.pre_attention_norm.build(input_shape)
        self.attention.build(input_shape)

        if self.use_post_attention_norm:
            shape = self.attention.compute_output_shape(input_shape)
            self.post_attention_norm.build(shape)

        shape = input_shape
        self.pre_ffw_norm.build(shape)
        self.gating_ffw.build(shape)
        self.gating_ffw_2.build(shape)

        shape = self.gating_ffw.compute_output_shape(shape)
        self.ffw_linear.build(shape)

        if self.use_post_ffw_norm:
            shape = self.ffw_linear.compute_output_shape(shape)
            self.post_ffw_norm.build(shape)

        self.built = True

    def compute_output_shape(self, input_shape):
        # Isometric
        return input_shape

    def _compute_image_bidirectional_attention_mask(self, vision_mask):
        # vision_mask is False for text, True for images. Shape of
        # (bsz, seq_len).
        bidirectional_mask = vision_mask

        # Left pad with 0.
        padded_mask = ops.cast(
            ops.pad(bidirectional_mask, [(0, 0), (1, 0)], constant_values=0),
            dtype="int32",
        )

        # Assign unique indices to every contiguous span of True.
        boundary = ops.cast(
            ops.greater(padded_mask[..., 1:], padded_mask[..., :-1]),
            dtype="int32",
        )
        numbered_boundary = ops.cumsum(boundary, -1)
        indices = ops.multiply(bidirectional_mask, numbered_boundary)

        indices_expanded_1 = ops.expand_dims(indices, 1)
        indices_expanded_2 = ops.expand_dims(indices, -1)

        mask = ops.logical_and(
            ops.equal(
                indices_expanded_1,
                indices_expanded_2,
            ),
            indices_expanded_2,
        )
        return mask

    def _compute_attention_mask(
        self,
        x,
        padding_mask,
        vision_mask,
        cache,
        cache_update_index,
    ):
        decoder_mask = merge_padding_and_attention_mask(
            inputs=x, padding_mask=padding_mask, attention_mask=None
        )

        batch_size = ops.shape(x)[0]
        input_length = output_length = ops.shape(x)[1]
        if cache is not None:
            input_length = ops.shape(cache)[2]

        if self.use_bidirectional_attention:
            # `output_length` and `input_length` will be the same in this case
            # because we use bidirectional attention for models like
            # `EmbeddingGemma` which aren't used for text generation.
            mask_1 = decoder_mask
            mask_2 = ops.transpose(mask_1, (0, 2, 1))
            return mask_1 * mask_2

        causal_mask = compute_causal_mask(
            batch_size=batch_size,
            input_length=input_length,
            output_length=output_length,
            cache_index=cache_update_index,
        )

        # Compute bidirectional mask (image tokens can attend to each other
        # in both directions, within the same image).
        if vision_mask is not None:
            bidirectional_image_mask = (
                self._compute_image_bidirectional_attention_mask(vision_mask)
            )
            causal_mask = ops.logical_or(causal_mask, bidirectional_image_mask)

        # Respect the padding mask.
        if decoder_mask is not None:
            causal_mask = ops.minimum(decoder_mask, causal_mask)

        return causal_mask

    def call(
        self,
        x,
        padding_mask=None,
        vision_mask=None,
        cache=None,
        cache_update_index=0,
        cache_update_mask=None,
    ):
        # Note: `vision_mask` is used only for Gemma3.
        normalized_x = self.pre_attention_norm(x)
        attention_mask = self._compute_attention_mask(
            normalized_x, padding_mask, vision_mask, cache, cache_update_index
        )
        if cache is not None:
            attention, new_cache = self.attention(
                normalized_x,
                attention_mask=attention_mask,
                cache=cache,
                cache_update_index=cache_update_index,
                cache_update_mask=cache_update_mask,
            )
        else:
            attention = self.attention(
                normalized_x,
                attention_mask=attention_mask,
            )

        if self.use_post_attention_norm:
            attention = self.post_attention_norm(attention)

        if self.dropout:
            attention = self.attention_dropout(attention)

        attention_x = x + attention
        normalized_x = self.pre_ffw_norm(attention_x)

        x1 = self.gating_ffw(normalized_x)
        x2 = self.gating_ffw_2(normalized_x)
        x = keras.activations.gelu(x1, approximate=True) * x2
        x = self.ffw_linear(x)

        if self.use_post_ffw_norm:
            x = self.post_ffw_norm(x)

        x = x + attention_x

        if cache is not None:
            return x, new_cache
        return x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "hidden_dim": self.hidden_dim,
                "intermediate_dim": self.intermediate_dim,
                "head_dim": self.head_dim,
                "num_query_heads": self.num_query_heads,
                "num_key_value_heads": self.num_key_value_heads,
                "query_head_dim_normalize": self.query_head_dim_normalize,
                "use_query_key_norm": self.use_query_key_norm,
                "use_post_ffw_norm": self.use_post_ffw_norm,
                "use_post_attention_norm": self.use_post_attention_norm,
                "gate_dim_reduction": self.gate_dim_reduction,
                "logit_soft_cap": self.logit_soft_cap,
                "use_sliding_window_attention": (
                    self.use_sliding_window_attention
                ),
                "sliding_window_size": self.sliding_window_size,
                "layer_norm_epsilon": self.layer_norm_epsilon,
                "dropout": self.dropout,
                "rope_wavelength": self.rope_wavelength,
                "rope_scaling_factor": self.rope_scaling_factor,
                "use_bidirectional_attention": self.use_bidirectional_attention,
            }
        )
        return config



import keras

from keras_hub.src.models.hgnetv2.hgnetv2_layers import HGNetV2Stage


@keras.saving.register_keras_serializable(package="keras_hub")
class HGNetV2Encoder(keras.layers.Layer):
    """This class represents the encoder of the HGNetV2 model.

    This class implements the encoder part of the HGNetV2 architecture, which
    consists of multiple stages. Each stage is an instance of `HGNetV2Stage`,
    and the encoder processes the input through these stages sequentially,
    collecting the hidden states at each stage.

    Args:
        stage_in_channels: A list of integers, specifying the input channels
            for each stage.
        stage_mid_channels: A list of integers, specifying the mid channels for
            each stage.
        stage_out_channels: A list of integers, specifying the output channels
            for each stage.
        stage_num_blocks: A list of integers, specifying the number of blocks
            in each stage.
        stage_num_of_layers: A list of integers, specifying the number of
            layers in each block of each stage.
        apply_downsample: A list of booleans or integers, indicating whether to
            downsample in each stage.
        use_lightweight_conv_block: A list of booleans, indicating whether to
            use HGNetV2 lightweight convolutional blocks in each stage.
        stage_kernel_size: A list of integers or tuples, specifying the kernel
            size for each stage.
        use_learnable_affine_block: A boolean, indicating whether to use
            learnable affine transformations in the blocks.
        data_format: `None` or str. If specified, either `"channels_last"` or
            `"channels_first"`. The ordering of the dimensions in the inputs.
            `"channels_last"` corresponds to inputs with shape
            `(batch_size, height, width, channels)` while `"channels_first"`
            corresponds to inputs with shape `(batch_size, channels, height,
            width)`. It defaults to the `image_data_format` value found in your
            Keras config file at `~/.keras/keras.json`. If you never set it,
            then it will be `"channels_last"`.
        channel_axis: int, the axis that represents the channels.
        **kwargs: Additional keyword arguments passed to the parent class.
    """

    def __init__(
        self,
        stage_in_channels,
        stage_mid_channels,
        stage_out_channels,
        stage_num_blocks,
        stage_num_of_layers,
        apply_downsample,
        use_lightweight_conv_block,
        stage_kernel_size,
        use_learnable_affine_block,
        data_format=None,
        channel_axis=None,
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.stage_in_channels = stage_in_channels
        self.stage_mid_channels = stage_mid_channels
        self.stage_out_channels = stage_out_channels
        self.stage_num_blocks = stage_num_blocks
        self.stage_num_of_layers = stage_num_of_layers
        self.apply_downsample = apply_downsample
        self.use_lightweight_conv_block = use_lightweight_conv_block
        self.stage_kernel_size = stage_kernel_size
        self.use_learnable_affine_block = use_learnable_affine_block
        self.data_format = data_format
        self.channel_axis = channel_axis

        self.stages_list = []
        for stage_idx in range(len(self.stage_in_channels)):
            stage_layer = HGNetV2Stage(
                stage_in_channels=self.stage_in_channels,
                stage_mid_channels=self.stage_mid_channels,
                stage_out_channels=self.stage_out_channels,
                stage_num_blocks=self.stage_num_blocks,
                stage_num_of_layers=self.stage_num_of_layers,
                apply_downsample=self.apply_downsample,
                use_lightweight_conv_block=self.use_lightweight_conv_block,
                stage_kernel_size=self.stage_kernel_size,
                use_learnable_affine_block=self.use_learnable_affine_block,
                stage_index=stage_idx,
                data_format=self.data_format,
                channel_axis=self.channel_axis,
                drop_path=0.0,
                name=f"{self.name}_stage_{stage_idx}"
                if self.name
                else f"stage_{stage_idx}",
                dtype=dtype,
            )
            self.stages_list.append(stage_layer)

    def build(self, input_shape):
        super().build(input_shape)
        current_input_shape = input_shape
        for stage_keras_layer in self.stages_list:
            stage_keras_layer.build(current_input_shape)
            current_input_shape = stage_keras_layer.compute_output_shape(
                current_input_shape
            )

    def call(
        self,
        hidden_state,
        training=None,
    ):
        all_hidden_states_list = []
        current_hidden_state = hidden_state
        for stage_keras_layer in self.stages_list:
            all_hidden_states_list.append(current_hidden_state)
            current_hidden_state = stage_keras_layer(
                current_hidden_state, training=training
            )
        all_hidden_states_list.append(current_hidden_state)
        return tuple(all_hidden_states_list)

    def compute_output_shape(self, input_shape):
        current_shape = input_shape
        all_hidden_shapes = [input_shape]
        for stage_keras_layer in self.stages_list:
            current_shape = stage_keras_layer.compute_output_shape(
                current_shape
            )
            all_hidden_shapes.append(current_shape)
        return tuple(all_hidden_shapes)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "stage_in_channels": self.stage_in_channels,
                "stage_mid_channels": self.stage_mid_channels,
                "stage_out_channels": self.stage_out_channels,
                "stage_num_blocks": self.stage_num_blocks,
                "stage_num_of_layers": self.stage_num_of_layers,
                "apply_downsample": self.apply_downsample,
                "use_lightweight_conv_block": self.use_lightweight_conv_block,
                "stage_kernel_size": self.stage_kernel_size,
                "use_learnable_affine_block": self.use_learnable_affine_block,
                "data_format": self.data_format,
                "channel_axis": self.channel_axis,
            }
        )
        return config



import keras

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.audio_to_text import AudioToText
from keras_hub.src.models.moonshine.moonshine_audio_to_text_preprocessor import (  # noqa: E501
    MoonshineAudioToTextPreprocessor,
)
from keras_hub.src.models.moonshine.moonshine_backbone import Arange
from keras_hub.src.models.moonshine.moonshine_backbone import MoonshineBackbone
from keras_hub.src.models.moonshine.moonshine_backbone import (
    compute_output_lengths,
)
from keras_hub.src.utils.tensor_utils import any_equal


@keras_hub_export("keras_hub.models.MoonshineAudioToText")
class MoonshineAudioToText(AudioToText):
    """An end-to-end Moonshine model for audio-to-text tasks.

    A Seq2Seq LM designed for audio-to-text tasks, such as speech recognition.
    The encoder processes audio features, and the decoder generates text
    transcriptions. You can finetune `MoonshineAudioToText` for any
    audio-to-text task (e.g., live transcription or voice commands).

    This model includes a `generate()` method for text generation based on audio
    inputs and an optional text prompt for the decoder. The generation strategy
    is controlled by a `sampler` argument passed to `compile()`. By default,
    `"top_k"` sampling is used.

    Args:
        backbone: A `keras_hub.models.MoonshineBackbone` instance.
        preprocessor: A `keras_hub.models.MoonshineAudioToTextPreprocessor` or
            `None`. If `None`, inputs must be preprocessed before calling the
            model.

    Examples:
    ```python
    # Initialize model from preset.
    moonshine_lm = keras_hub.models.MoonshineAudioToText.from_preset(
        "moonshine_base"
    )

    # Generate with single audio input.
    audio_tensor = keras.random.normal((1, 16000, 1))
    moonshine_lm.generate({"audio": audio_tensor})

    # Generate with text prompt.
    moonshine_lm.generate({"audio": audio_tensor, "text": "quick"})

    # Use different sampling strategy.
    moonshine_lm.compile(sampler="greedy")
    moonshine_lm.generate({"audio": audio_tensor})
    ```
    """

    # References:
    # Defined and formulated based on the Hugging Face implementation of the
    # MoonshineForConditionalGeneration class (https://github.com/huggingface/transformers/blob/dcbdf7e962c4b36140cc9ee76f870016121e69e5/src/transformers/models/moonshine/modeling_moonshine.py#L1509-L1626).

    backbone_cls = MoonshineBackbone
    preprocessor_cls = MoonshineAudioToTextPreprocessor

    def __init__(self, backbone, preprocessor=None, **kwargs):
        # === Layers ===
        self.backbone = backbone
        self.preprocessor = preprocessor

        # === Functional Model ===
        inputs = backbone.input
        hidden_states = backbone(inputs)["decoder_sequence_output"]
        outputs = backbone.token_embedding(hidden_states, reverse=True)
        super().__init__(
            inputs=inputs,
            outputs=outputs,
            **kwargs,
        )

    def call_decoder_with_cache(
        self,
        encoder_hidden_states,
        encoder_padding_mask,
        decoder_token_ids,
        self_attention_cache=None,
        self_attention_cache_update_index=None,
        cross_attention_cache=None,
    ):
        """Process decoder inputs with attention caching for efficient
        generation.

        Args:
            encoder_hidden_states: Tensor. Encoder outputs.
            encoder_padding_mask: Tensor. Padding mask for encoder outputs.
            decoder_token_ids: Tensor. Decoder input token IDs.
            self_attention_cache: Tensor, optional. Cache for self-attention
                layers.
            self_attention_cache_update_index: int, optional. Index for cache
                updates.
            cross_attention_cache: Tensor, optional. Cache for cross-attention
                layers. This cache is computed once and reused.

        Returns:
            Tuple: Tuple of (logits, hidden_states, new_self_attention_cache,
            cross_attention_cache).
        """
        tokens = self.backbone.token_embedding(decoder_token_ids)
        x = tokens

        # Cache management for audio-to-text generation.
        self_attention_caches = []
        position = keras.ops.array(
            [self_attention_cache_update_index], dtype="int32"
        )
        rotary_embedding = self.backbone.decoder_rotary_embedding(position)

        for i, layer in enumerate(self.backbone.decoder_blocks):
            current_self_cache = self_attention_cache[:, i, ...]
            current_cross_cache = cross_attention_cache[:, i, ...]
            x, new_self_cache = layer(
                decoder_sequence=x,
                encoder_sequence=encoder_hidden_states,
                rotary_embedding=rotary_embedding,
                encoder_padding_mask=encoder_padding_mask,
                self_attention_cache=current_self_cache,
                self_attention_cache_update_index=self_attention_cache_update_index,
                cross_attention_cache=current_cross_cache,
                training=False,
            )
            # Update self-attention cache.
            self_attention_caches.append(new_self_cache)

        # [batch_size, num_layers, 2, seq_len, num_heads, head_dim].
        new_self_attention_cache = keras.ops.stack(
            self_attention_caches, axis=1
        )
        hidden_states = self.backbone.decoder_post_norm(x)
        logits = self.backbone.token_embedding(hidden_states, reverse=True)
        return (
            logits,
            hidden_states,
            new_self_attention_cache,
            cross_attention_cache,
        )

    def _build_cache(
        self,
        encoder_input_values,
        encoder_padding_mask,
        decoder_token_ids,
        decoder_padding_mask,
    ):
        """Build initial cache states from inputs."""
        encoder_hidden_states, encoder_attention_mask_for_decoder = (
            self.call_encoder(
                encoder_input_values=encoder_input_values,
                padding_mask=encoder_padding_mask,
            )
        )
        precomputed_cross_caches = []
        for layer in self.backbone.decoder_blocks:
            cross_k = layer.cross_attention._key_dense(encoder_hidden_states)
            cross_v = layer.cross_attention._value_dense(encoder_hidden_states)
            layer_cross_cache = keras.ops.stack([cross_k, cross_v], axis=1)
            precomputed_cross_caches.append(layer_cross_cache)
        precomputed_cross_cache = keras.ops.stack(
            precomputed_cross_caches, axis=1
        )
        batch_size = keras.ops.shape(encoder_input_values)[0]
        num_layers = self.backbone.decoder_num_layers
        num_heads = self.backbone.decoder_num_heads
        head_dim = self.backbone.hidden_dim // self.backbone.decoder_num_heads
        if self.backbone.pad_head_dim_to_multiple_of is not None:
            head_dim = (
                (head_dim + self.backbone.pad_head_dim_to_multiple_of - 1)
                // self.backbone.pad_head_dim_to_multiple_of
            ) * self.backbone.pad_head_dim_to_multiple_of
        # Use the full sequence length for the cache dimension.
        cache_length = keras.ops.shape(decoder_token_ids)[1]
        initial_self_cache_shape = (
            batch_size,
            num_layers,
            2,
            cache_length,
            num_heads,
            head_dim,
        )
        initial_self_cache = keras.ops.zeros(
            initial_self_cache_shape, dtype=self.compute_dtype
        )
        tokens = self.backbone.token_embedding(decoder_token_ids)
        x = tokens
        positions = keras.ops.arange(0, cache_length, dtype="int32")
        rotary_embedding = self.backbone.decoder_rotary_embedding(positions)
        seeded_self_caches = []
        for i, layer in enumerate(self.backbone.decoder_blocks):
            current_initial_self_cache = initial_self_cache[:, i, ...]
            current_precomputed_cross_cache = precomputed_cross_cache[:, i, ...]
            x, seeded_self_cache_layer = layer(
                decoder_sequence=x,
                encoder_sequence=encoder_hidden_states,
                rotary_embedding=rotary_embedding,
                decoder_padding_mask=decoder_padding_mask,
                encoder_padding_mask=encoder_attention_mask_for_decoder,
                self_attention_cache=current_initial_self_cache,
                self_attention_cache_update_index=0,
                cross_attention_cache=current_precomputed_cross_cache,
                training=False,
            )
            seeded_self_caches.append(seeded_self_cache_layer)
        hidden_states = self.backbone.decoder_post_norm(x)
        self_attn_cache = keras.ops.stack(seeded_self_caches, axis=1)
        return (
            hidden_states,
            self_attn_cache,
            precomputed_cross_cache,
            encoder_hidden_states,
            encoder_attention_mask_for_decoder,
        )

    def call_encoder(self, encoder_input_values, padding_mask):
        """Process audio input through the encoder stack."""
        x = self.backbone.conv1(encoder_input_values)
        x = self.backbone.tanh_after_conv1(x)
        x = self.backbone.group_norm(x)
        x = self.backbone.conv2(x)
        x = self.backbone.gelu_after_conv2(x)
        x = self.backbone.conv3(x)
        x = self.backbone.gelu_after_conv3(x)
        original_lengths = keras.ops.sum(
            keras.ops.cast(padding_mask, "int32"), axis=1
        )
        output_lengths = compute_output_lengths(original_lengths)
        padding_mask = self.backbone._compute_mask_layer(x, output_lengths)
        positions = Arange(name="encoder_positions")(x)
        rotary_embedding = self.backbone.encoder_rotary_embedding(positions)
        x = self.backbone.encoder_dropout(x, training=False)
        for transformer_layer in self.backbone.encoder_blocks:
            x = transformer_layer(
                inputs=x,
                rotary_embedding=rotary_embedding,
                attention_mask=padding_mask,
                training=False,
            )
        x = self.backbone.encoder_final_layer_norm(x)
        return x, padding_mask

    # Source: https://github.com/huggingface/transformers/blob/9e94801146ceeb3b215bbdb9492be74d7d7b7210/src/transformers/generation/utils.py#L1970-L2463
    def generate_step(self, inputs, stop_token_ids=None):
        """A compilable generation function for a batch of inputs.

        This function represents the inner, XLA-compilable, generation function
        for a single batch of inputs. Inputs should have the same structure as
        model inputs, a dictionary with keys `"encoder_input_values"`,
        `"encoder_padding_mask"`, `"decoder_token_ids"` and
        `"decoder_padding_mask"`.

        Args:
            inputs: A dictionary with four keys - `"encoder_input_values"`,
                `"encoder_padding_mask"`, `"decoder_token_ids"` and
                `"decoder_padding_mask"`, with batched tensor values.
            stop_token_ids: Tuple of id's of end token's to stop on. If all
                sequences have produced a new stop token, generation
                will stop.

        Returns:
            Dictionary: A dictionary with two keys - `"decoder_token_ids"`
                containing the updated token sequence with newly generated
                tokens, and `"decoder_padding_mask"` containing the updated
                padding mask for the generated sequence.
        """
        encoder_input_values = inputs["encoder_input_values"]
        encoder_padding_mask = inputs["encoder_padding_mask"]
        decoder_token_ids = inputs["decoder_token_ids"]
        decoder_padding_mask = inputs["decoder_padding_mask"]

        if (
            encoder_input_values is None
            or encoder_padding_mask is None
            or decoder_token_ids is None
        ):
            raise ValueError("Input tensors cannot be None")

        (
            hidden_states,
            self_attention_cache,
            cross_attention_cache,
            encoder_hidden_states,
            encoder_attention_mask_for_decoder,
        ) = self._build_cache(
            encoder_input_values,
            encoder_padding_mask,
            decoder_token_ids,
            decoder_padding_mask,
        )
        row_lengths = keras.ops.sum(
            keras.ops.cast(decoder_padding_mask, "int32"),
            axis=-1,
        )
        index = keras.ops.min(row_lengths)

        def next(prompt, cache, index):
            if isinstance(cache, tuple) and len(cache) == 2:
                current_self_attention_cache = cache[0]
                current_cross_attention_cache = cache[1]
            elif cache is not None and not isinstance(cache, tuple):
                current_self_attention_cache = cache
                current_cross_attention_cache = cross_attention_cache
            else:
                cache = None
            cache_index = index - 1
            num_samples = keras.ops.shape(prompt)[0]
            next_token_input = keras.ops.slice(
                prompt, [0, cache_index], [num_samples, 1]
            )

            batch_size = keras.ops.shape(encoder_input_values)[0]

            def repeat_tensor(x):
                if keras.ops.shape(x)[0] == num_samples:
                    return x
                return keras.ops.repeat(
                    x, repeats=num_samples // batch_size, axis=0
                )

            cross_attention_cache_repeated = repeat_tensor(
                current_cross_attention_cache
            )
            logits, hidden_states, new_self_attention_cache, _ = (
                self.call_decoder_with_cache(
                    encoder_hidden_states=repeat_tensor(encoder_hidden_states),
                    encoder_padding_mask=repeat_tensor(
                        encoder_attention_mask_for_decoder
                    ),
                    decoder_token_ids=next_token_input,
                    self_attention_cache=current_self_attention_cache,
                    self_attention_cache_update_index=cache_index,
                    cross_attention_cache=cross_attention_cache_repeated,
                )
            )
            return (
                logits[:, 0, :],
                hidden_states[:, 0, :],
                (new_self_attention_cache, current_cross_attention_cache),
            )

        decoder_token_ids = self.sampler(
            next=next,
            prompt=decoder_token_ids,
            cache=(self_attention_cache, cross_attention_cache),
            index=index,
            mask=keras.ops.cast(
                decoder_token_ids != self.preprocessor.tokenizer.pad_token_id
                if self.preprocessor is not None
                else decoder_padding_mask,
                dtype="bool",
            ),
            stop_token_ids=stop_token_ids,
            hidden_states=hidden_states,
            model=self,
        )

        if stop_token_ids is not None:
            end_locations = any_equal(
                decoder_token_ids,
                stop_token_ids,
                decoder_token_ids == self.preprocessor.tokenizer.pad_token_id
                if self.preprocessor is not None
                else False,
            )
            end_locations = keras.ops.cast(end_locations, "int32")
            cumsum = keras.ops.cumsum(end_locations, axis=-1)
            overflow = cumsum - end_locations
            decoder_padding_mask = keras.ops.logical_not(
                keras.ops.cast(overflow, "bool")
            )
        else:
            decoder_padding_mask = keras.ops.ones_like(
                decoder_token_ids, dtype="bool"
            )

        return {
            "decoder_token_ids": decoder_token_ids,
            "decoder_padding_mask": decoder_padding_mask,
        }



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.image_classifier_preprocessor import (
    ImageClassifierPreprocessor,
)
from keras_hub.src.models.vgg.vgg_backbone import VGGBackbone
from keras_hub.src.models.vgg.vgg_image_converter import VGGImageConverter


@keras_hub_export("keras_hub.models.VGGImageClassifierPreprocessor")
class VGGImageClassifierPreprocessor(ImageClassifierPreprocessor):
    backbone_cls = VGGBackbone
    image_converter_cls = VGGImageConverter



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.opt.opt_backbone import OPTBackbone
from keras_hub.src.tokenizers.byte_pair_tokenizer import BytePairTokenizer


@keras_hub_export(
    [
        "keras_hub.tokenizers.OPTTokenizer",
        "keras_hub.models.OPTTokenizer",
    ]
)
class OPTTokenizer(BytePairTokenizer):
    """An OPT tokenizer using Byte-Pair Encoding subword segmentation.

    This tokenizer class will tokenize raw strings into integer sequences and
    is based on `keras_hub.tokenizers.BytePairTokenizer`. Unlike the
    underlying tokenizer, it will check for all special tokens needed by OPT
    models and provides a `from_preset()` method to automatically download
    a matching vocabulary for a OPT preset.

    If input is a batch of strings (rank > 0), the layer will output a
    `tf.RaggedTensor` where the last dimension of the output is ragged.
    If input is a scalar string (rank == 0), the layer will output a dense
    `tf.Tensor` with static shape `[None]`.

    Args:
        vocabulary: string or dict, maps token to integer ids. If it is a
            string, it should be the file path to a json file.
        merges: string or list, contains the merge rule. If it is a string,
            it should be the file path to merge rules. The merge rule file
            should have one merge rule per line. Every merge rule contains
            merge entities separated by a space.

    Examples:
    ```python
    # Unbatched input.
    tokenizer = keras_hub.models.OPTTokenizer.from_preset(
        "opt_125m_en",
    )
    tokenizer("The quick brown fox jumped.")

    # Batched input.
    tokenizer(["The quick brown fox jumped.", "The fox slept."])

    # Detokenization.
    tokenizer.detokenize(tokenizer("The quick brown fox jumped."))

    # Custom vocabulary.
    vocab = {"<pad>": 1, "</s>": 2, "Ġquick": 4, "Ġfox": 5}
    merges = ["Ġ q", "u i", "c k", "ui ck", "Ġq uick"]
    merges += ["Ġ f", "o x", "Ġf ox"]
    tokenizer = keras_hub.models.OPTTokenizer(vocabulary=vocab, merges=merges)
    tokenizer("The quick brown fox jumped.")
    ```
    """

    backbone_cls = OPTBackbone

    def __init__(
        self,
        vocabulary=None,
        merges=None,
        **kwargs,
    ):
        self._add_special_token("</s>", "end_token")
        self._add_special_token("</s>", "start_token")
        self._add_special_token("<pad>", "pad_token")
        super().__init__(
            vocabulary=vocabulary,
            merges=merges,
            **kwargs,
        )



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.llama3.llama3_backbone import Llama3Backbone
from keras_hub.src.tokenizers.byte_pair_tokenizer import BytePairTokenizer


@keras_hub_export(
    [
        "keras_hub.tokenizers.Llama3Tokenizer",
        "keras_hub.models.Llama3Tokenizer",
    ]
)
class Llama3Tokenizer(BytePairTokenizer):
    backbone_cls = Llama3Backbone

    def __init__(
        self,
        vocabulary=None,
        merges=None,
        bos_token="<|begin_of_text|>",
        eos_token="<|end_of_text|>",
        misc_special_tokens={"<|start_header_id|>", "<|end_header_id|>"},
        **kwargs,
    ):
        # Note: all special tokens must also appear in "vocabulary"

        self._add_special_token(bos_token, "start_token")
        misc_special_tokens -= {bos_token}
        self._add_special_token(eos_token, "end_token")
        misc_special_tokens -= {eos_token}
        for i, token in enumerate(misc_special_tokens):
            self._add_special_token(token, f"special_token_{i:03d}")

        # Hack:
        # Llama models use the <|end_of_text|> or the <|eot_id|> as the stop
        # token. This info can be read from config when loading a Hugging Face
        # checkpoint but no such config exists for Keras checkpoints.
        # Setting both probable end tokens when no config is availble will
        # make text generation work in all cases as it will stop
        # on both end tokens. However, the packer will always use
        # "<|end_of_text|>" , which will be the wrong eos_token for "instruct"
        # variants of Llama3.
        # TODO: load this correctly from a Keras tokenizer config.
        if eos_token == "<|end_of_text|>":
            self._add_special_token("<|eot_id|>", "end_token2")

        self.pad_token_id = 0
        super().__init__(
            vocabulary=vocabulary,
            merges=merges,
            **kwargs,
        )



from keras import layers

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.dinov2.dinov2_layers import DINOV2Embedding
from keras_hub.src.models.dinov2.dinov2_layers import DINOV2Encoder
from keras_hub.src.models.feature_pyramid_backbone import FeaturePyramidBackbone
from keras_hub.src.utils.keras_utils import standardize_data_format


@keras_hub_export("keras_hub.models.DINOV2Backbone")
class DINOV2Backbone(FeaturePyramidBackbone):
    """DINOV2 core network with hyperparameters.

    DINOV2 offers a powerful, generalist visual backbone learned entirely from
    unlabeled images as described in [DINOv2: Learning Robust Visual Features
    without Supervision](https://arxiv.org/abs/2304.07193)

    The default constructor gives a fully customizable, randomly initialized
    DINOV2 model with any number of layers, heads, and embedding dimensions. To
    load preset architectures and weights, use the `from_preset` constructor.

    Note that this backbone is a Feature Pyramid Backbone that can output
    intermediate feature maps from different stages of the model. See the
    example below for how to access these feature pyramid outputs.

    Note that this backbone supports interpolation of the position embeddings
    to the input image shape. This is useful when the input image shape is
    different from the shape used to train the position embeddings. The
    `position_embedding_shape` argument is used to specify the original shape
    used to train the position embeddings.

    Args:
        patch_size: int. The size of each square patch in the input image.
        num_layers: int. The number of transformer layers.
        hidden_dim: int. The size of the transformer hidden state at the end
            of each transformer layer.
        num_heads: int. The number of attention heads for each transformer.
        intermediate_dim: int. The output dimension of the first Dense layer in
            a two-layer feedforward network for each transformer.
        layer_scale_init_value: float. The initial value for the layer scale in
            the transformer layers. Defaults to `1.0`.
        num_register_tokens: int. The number of register tokens to use in the
            embedding layer. Defaults to `0`.
        use_mask_token: bool. Whether to use a mask token in the embedding
            layer. Defaults to `True`.
        use_swiglu_ffn: bool. Whether to use SwigLU activation in the MLP
            layers. Defaults to `False`.
        dropout_rate: float. The dropout rate to use. Defaults to `0.0`.
        drop_path_rate: float. The drop path rate to use. Defaults to `0.0`.
        image_shape: tuple. The input shape without the batch size. Defaults to
            `(224, 224, 3)`.
        position_embedding_shape: tuple. The original shape used to train the
            position embeddings. This is used to interpolate the position
            embeddings to the actual input shape. Defaults to `(518, 518)`.
        antialias_in_interpolation: bool. Whether to use antialiasing in the
            interpolation of the position embeddings. Defaults to `False`.
        apply_layernorm: bool. Whether to apply layer normalization to the
            outputs of each stage in the feature pyramid. Defaults to `False`.
        data_format: `None` or str. If specified, either `"channels_last"` or
            `"channels_first"`. The ordering of the dimensions in the
            inputs. `"channels_last"` corresponds to inputs with shape
            `(batch_size, height, width, channels)`
            while `"channels_first"` corresponds to inputs with shape
            `(batch_size, channels, height, width)`. It defaults to the
            `image_data_format` value found in your Keras config file at
            `~/.keras/keras.json`. If you never set it, then it will be
            `"channels_last"`.
        dtype: string or `keras.mixed_precision.DTypePolicy`. The dtype to use
            for the models computations and weights. Note that some
            computations, such as softmax and layer normalization will always
            be done a float32 precision regardless of dtype.

    Example:
    ```python
    # Pretrained DINOV2 model.
    input_data = {
        "images": np.ones(shape=(1, 518, 518, 3), dtype="float32"),
    }
    model = keras_hub.models.DINOV2Backbone.from_preset(
        "dinov2_base"
    )
    model(input_data)

    # Pretrained DINOV2 model with custom image shape.
    input_data = {
        "images": np.ones(shape=(1, 224, 224, 3), dtype="float32"),
    }
    model = keras_hub.models.DINOV2Backbone.from_preset(
        "dinov2_base", image_shape=(224, 224, 3)
    )
    model(input_data)

    # Randomly initialized DINOV2 model with custom config.
    model = keras_hub.models.DINOV2Backbone(
        patch_size=14,
        num_layers=2,
        hidden_dim=32,
        num_heads=2,
        intermediate_dim=128,
        image_shape=(224, 224, 3),
        position_embedding_shape=(518, 518),
    )
    model(input_data)

    # Accessing feature pyramid outputs.
    backbone = keras_hub.models.DINOV2Backbone.from_preset(
        "dinov2_base", image_shape=(224, 224, 3)
    )
    model = keras.Model(
        inputs=backbone.inputs,
        outputs=backbone.pyramid_outputs,
    )
    features = model(input_data)
    ```
    """

    def __init__(
        self,
        patch_size,
        num_layers,
        hidden_dim,
        num_heads,
        intermediate_dim,
        layer_scale_init_value=1.0,
        num_register_tokens=0,
        use_mask_token=True,
        use_swiglu_ffn=False,
        dropout_rate=0.0,
        drop_path_rate=0.0,
        image_shape=(224, 224, 3),
        position_embedding_shape=(518, 518, 3),
        antialias_in_interpolation=False,
        apply_layernorm=False,
        data_format=None,
        dtype=None,
        name=None,
        **kwargs,
    ):
        data_format = standardize_data_format(data_format)
        if data_format == "channels_last":
            height, width = image_shape[0], image_shape[1]
            position_embedding_height, position_embedding_width = (
                position_embedding_shape[0],
                position_embedding_shape[1],
            )
        else:
            height, width = image_shape[1], image_shape[2]
            position_embedding_height, position_embedding_width = (
                position_embedding_shape[1],
                position_embedding_shape[2],
            )
        if height != width:
            raise ValueError(
                "`DINOV2Backbone` expects the height and width to be the "
                f"same in `image_shape`. Received: image_shape={image_shape}"
            )

        # `prefix` is used to prevent duplicate name when utilizing multiple
        # DINOV2Backbone encoders within a single model.
        prefix = str(name) + "_" if name is not None else ""

        # === Layers ===
        self.embeddings = DINOV2Embedding(
            hidden_dim=hidden_dim,
            patch_size=patch_size,
            image_shape=(height, width),
            num_register_tokens=num_register_tokens,
            use_mask_token=use_mask_token,
            dropout_rate=dropout_rate,
            position_embedding_shape=(
                position_embedding_height,
                position_embedding_width,
            ),
            antialias_in_interpolation=antialias_in_interpolation,
            data_format=data_format,
            dtype=dtype,
            name=f"{prefix}embeddings",
        )
        self.encoder = DINOV2Encoder(
            num_layers=num_layers,
            hidden_dim=hidden_dim,
            num_heads=num_heads,
            intermediate_dim=intermediate_dim,
            layer_scale_init_value=layer_scale_init_value,
            use_swiglu_ffn=use_swiglu_ffn,
            dropout_rate=dropout_rate,
            drop_path_rate=drop_path_rate,
            dtype=dtype,
            name=f"{prefix}encoder",
        )
        self.layernorm = layers.LayerNormalization(
            epsilon=1e-6, dtype=dtype, name=f"{prefix}layernorm"
        )

        # === Functional Model ===
        pyramid_outputs = {}
        image_input = layers.Input(shape=image_shape, name="images")
        x = self.embeddings(image_input)
        pyramid_outputs["stem"] = x
        x, encoder_pyramid_outputs = self.encoder(x)
        pyramid_outputs.update(encoder_pyramid_outputs)
        x = self.layernorm(x)
        if apply_layernorm:
            for key in pyramid_outputs:
                pyramid_outputs[key] = self.layernorm(pyramid_outputs[key])
        outputs = x
        super().__init__(
            inputs={"images": image_input},
            outputs=outputs,
            dtype=dtype,
            name=name,
            **kwargs,
        )

        # === Config ===
        self.patch_size = int(patch_size)
        self.num_layers = int(num_layers)
        self.hidden_dim = int(hidden_dim)
        self.num_heads = int(num_heads)
        self.intermediate_dim = int(intermediate_dim)
        self.layer_scale_init_value = float(layer_scale_init_value)
        self.num_register_tokens = int(num_register_tokens)
        self.use_mask_token = bool(use_mask_token)
        self.use_swiglu_ffn = bool(use_swiglu_ffn)
        self.dropout_rate = float(dropout_rate)
        self.drop_path_rate = float(drop_path_rate)
        self.image_shape = image_shape
        self.position_embedding_shape = position_embedding_shape
        self.antialias_in_interpolation = bool(antialias_in_interpolation)
        self.apply_layernorm = apply_layernorm
        self.pyramid_outputs = pyramid_outputs

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "patch_size": self.patch_size,
                "num_layers": self.num_layers,
                "hidden_dim": self.hidden_dim,
                "num_heads": self.num_heads,
                "intermediate_dim": self.intermediate_dim,
                "layer_scale_init_value": self.layer_scale_init_value,
                "num_register_tokens": self.num_register_tokens,
                "use_mask_token": self.use_mask_token,
                "use_swiglu_ffn": self.use_swiglu_ffn,
                "dropout_rate": self.dropout_rate,
                "drop_path_rate": self.drop_path_rate,
                "image_shape": self.image_shape,
                "position_embedding_shape": self.position_embedding_shape,
                "antialias_in_interpolation": self.antialias_in_interpolation,
                "apply_layernorm": self.apply_layernorm,
            }
        )
        return config



import keras

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.backbone import Backbone
from keras_hub.src.models.mobilenet.util import adjust_channels

BN_EPSILON = 1e-5
BN_MOMENTUM = 0.9


class SqueezeAndExcite2D(keras.layers.Layer):
    """
    Description:
        This layer applies a content-aware mechanism to adaptively assign
        channel-wise weights. It uses global average pooling to compress
        feature maps into single values, which are then processed by
        two Conv1D layers: the first reduces the dimensionality, and
        the second restores it.
    Args:
        filters: Number of input and output filters. The number of input and
            output filters is same.
        bottleneck_filters: (Optional) Number of bottleneck filters. Defaults
            to `0.25 * filters`
        squeeze_activation: (Optional) String, callable (or
            keras.layers.Layer) or keras.activations.Activation instance
            denoting activation to be applied after squeeze convolution.
            Defaults to `relu`.
        excite_activation: (Optional) String, callable (or
            keras.layers.Layer) or keras.activations.Activation instance
            denoting activation to be applied after excite convolution.
            Defaults to `sigmoid`.
        name: Name of the layer
        dtype: `None` or str or `keras.mixed_precision.DTypePolicy`. The dtype
            to use for the model's computations and weights.
    """

    def __init__(
        self,
        filters,
        bottleneck_filters=None,
        squeeze_activation="relu",
        excite_activation="sigmoid",
        name=None,
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.filters = filters
        self.bottleneck_filters = bottleneck_filters
        self.squeeze_activation = squeeze_activation
        self.excite_activation = excite_activation
        self.name = name

        image_data_format = keras.config.image_data_format()
        if image_data_format == "channels_last":
            self.spatial_dims = (1, 2)
        else:
            self.spatial_dims = (2, 3)

        self.conv_reduce = keras.layers.Conv2D(
            bottleneck_filters,
            (1, 1),
            data_format=image_data_format,
            name=f"{name}_conv_reduce",
            dtype=dtype,
        )
        self.activation1 = keras.layers.Activation(
            self.squeeze_activation,
            name=self.name + "squeeze_activation",
            dtype=dtype,
        )

        self.conv_expand = keras.layers.Conv2D(
            filters,
            (1, 1),
            data_format=image_data_format,
            name=f"{name}_conv_expand",
            dtype=dtype,
        )
        self.gate = keras.layers.Activation(
            self.excite_activation,
            name=self.name + "excite_activation",
            dtype=dtype,
        )

    def compute_output_shape(self, input_shape):
        shape = self.conv_reduce.compute_output_shape(input_shape)
        shape = self.activation1.compute_output_shape(shape)
        shape = self.conv_expand.compute_output_shape(shape)
        return self.gate.compute_output_shape(shape)

    def build(self, input_shape):
        self.conv_reduce.build(input_shape)
        input_shape = self.conv_reduce.compute_output_shape(input_shape)
        self.activation1.build(input_shape)
        input_shape = self.activation1.compute_output_shape(input_shape)
        self.conv_expand.build(input_shape)
        input_shape = self.conv_expand.compute_output_shape(input_shape)
        self.gate.build(input_shape)
        self.built = True

    def call(self, inputs):
        x_se = keras.ops.mean(inputs, axis=self.spatial_dims, keepdims=True)
        x_se = self.conv_reduce(x_se)
        x_se = self.activation1(x_se)
        x_se = self.conv_expand(x_se)
        return inputs * self.gate(x_se)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "filters": self.filters,
                "bottleneck_filters": self.bottleneck_filters,
                "squeeze_activation": self.squeeze_activation,
                "excite_activation": self.excite_activation,
                "name": self.name,
                "spatial_dims": self.spatial_dims,
            }
        )
        return config


class DepthwiseConvBlock(keras.layers.Layer):
    """
    A depthwise convolution block consists of a depthwise conv,
    batch normalization, relu, optional squeeze & excite, pointwise convolution,
    and batch normalization layer.

    Args:
        infilters: int, the output channels for the initial depthwise conv
        filters: int, the dimensionality of the output space
            (i.e. the number of output filters in the pointwise convolution).
        kernel_size: int or Tuple[int, int], the kernel size to apply
            to the initial depthwise convolution
        strides: An int or Tuple[int, int], specifying the strides
            of the convolution along the width and height.
            Can be a single integer to specify the same value for
            all spatial dimensions.
        squeeze_excite_ratio: squeeze & excite ratio: float[Optional], if
            exists, specifies the ratio of channels (<1) to squeeze the initial
            signal into before reexciting back out. If (>1) technically, it's an
            excite & squeeze layer. If this doesn't exist there is no
            SqueezeExcite layer.
        residual: bool, default False. True if we want a residual connection. If
            False, there is no residual connection.
        name: str, name of the layer
        dtype: `None` or str or `keras.mixed_precision.DTypePolicy`. The dtype
            to use for the model's computations and weights.

    Input shape when applied as a layer:
        4D tensor with shape: `(batch, rows, cols, channels)` in "channels_last"
        4D tensor with shape: `(batch, channels, rows, cols)` in
            "channels_first"
    Returns:
        Output tensor of block.
    """

    def __init__(
        self,
        infilters,
        filters,
        kernel_size=3,
        stride=2,
        squeeze_excite_ratio=None,
        residual=False,
        name=None,
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.infilters = infilters
        self.filters = filters
        self.kernel_size = kernel_size
        self.stride = stride
        self.squeeze_excite_ratio = squeeze_excite_ratio
        self.residual = residual
        self.name = name

        channel_axis = (
            -1 if keras.config.image_data_format() == "channels_last" else 1
        )
        self.name = name = f"{name}_0"

        self.pad = keras.layers.ZeroPadding2D(
            padding=(1, 1),
            name=f"{name}_pad",
            dtype=dtype,
        )
        self.conv1 = keras.layers.Conv2D(
            infilters,
            kernel_size,
            strides=stride,
            padding="valid",
            data_format=keras.config.image_data_format(),
            groups=infilters,
            use_bias=False,
            name=f"{name}_conv1",
            dtype=dtype,
        )
        self.batch_normalization1 = keras.layers.BatchNormalization(
            axis=channel_axis,
            epsilon=BN_EPSILON,
            momentum=BN_MOMENTUM,
            name=f"{name}_bn1",
            dtype=dtype,
        )
        self.activation1 = keras.layers.ReLU(dtype=dtype)

        if squeeze_excite_ratio:
            self.se_layer = SqueezeAndExcite2D(
                filters=infilters,
                bottleneck_filters=adjust_channels(
                    infilters * squeeze_excite_ratio
                ),
                squeeze_activation="relu",
                excite_activation=keras.activations.hard_sigmoid,
                name=f"{name}_squeeze_excite",
                dtype=dtype,
            )

        self.conv2 = keras.layers.Conv2D(
            filters,
            kernel_size=1,
            data_format=keras.config.image_data_format(),
            use_bias=False,
            name=f"{name}_conv2",
            dtype=dtype,
        )
        self.batch_normalization2 = keras.layers.BatchNormalization(
            axis=channel_axis,
            epsilon=BN_EPSILON,
            momentum=BN_MOMENTUM,
            name=f"{name}_bn2",
            dtype=dtype,
        )

    def build(self, input_shape):
        self.pad.build(input_shape)
        input_shape = self.pad.compute_output_shape(input_shape)
        self.conv1.build(input_shape)
        input_shape = self.conv1.compute_output_shape(input_shape)
        self.batch_normalization1.build(input_shape)
        input_shape = self.batch_normalization1.compute_output_shape(
            input_shape
        )
        self.activation1.build(input_shape)
        input_shape = self.activation1.compute_output_shape(input_shape)
        if self.squeeze_excite_ratio:
            self.se_layer.build(input_shape)
            input_shape = self.se_layer.compute_output_shape(input_shape)
        self.conv2.build(input_shape)
        input_shape = self.conv2.compute_output_shape(input_shape)
        self.batch_normalization2.build(input_shape)
        self.built = True

    def call(self, inputs):
        x = self.pad(inputs)
        x = self.conv1(x)
        x = self.batch_normalization1(x)
        x = self.activation1(x)

        if self.squeeze_excite_ratio:
            x = self.se_layer(x)

        x = self.conv2(x)
        x = self.batch_normalization2(x)

        if self.residual:
            x = x + inputs

        return x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "infilters": self.infilters,
                "filters": self.filters,
                "kernel_size": self.kernel_size,
                "stride": self.stride,
                "squeeze_excite_ratio": self.squeeze_excite_ratio,
                "residual": self.residual,
                "name": self.name,
            }
        )
        return config


class InvertedResidualBlock(keras.layers.Layer):
    """An Inverted Residual Block.

    Args:
        expansion: integer, the expansion ratio, multiplied with infilters to
            get the minimum value passed to adjust_channels.
        infilters: Int, the output channels for the initial depthwise conv
        filters: integer, number of filters for convolution layer.
        kernel_size: integer, the kernel size for DepthWise Convolutions.
        stride: integer, the stride length for DepthWise Convolutions.
        squeeze_excite_ratio: float, ratio for bottleneck filters. Number of
            bottleneck filters = filters * se_ratio.
        activation: the activation layer to use.
        padding: padding in the conv2d layer
        name: string, block label.
        dtype: `None` or str or `keras.mixed_precision.DTypePolicy`. The dtype
            to use for the model's computations and weights.

    Input shape when applied as a layer:
        4D tensor with shape: `(batch, rows, cols, channels)` in "channels_last"
        4D tensor with shape: `(batch, channels, rows, cols)` in
            "channels_first"
    Returns:
        Output tensor of block.
    """

    def __init__(
        self,
        expansion,
        infilters,
        filters,
        kernel_size,
        stride,
        squeeze_excite_ratio,
        activation,
        padding,
        name=None,
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.expansion = expansion
        self.infilters = infilters
        self.filters = filters
        self.kernel_size = kernel_size
        self.stride = stride
        self.squeeze_excite_ratio = squeeze_excite_ratio
        self.activation = activation
        self.padding = padding
        self.name = name

        channel_axis = (
            -1 if keras.config.image_data_format() == "channels_last" else 1
        )
        expanded_channels = adjust_channels(expansion)

        self.conv1 = keras.layers.Conv2D(
            expanded_channels,
            kernel_size=1,
            data_format=keras.config.image_data_format(),
            use_bias=False,
            name=f"{name}_conv1",
            dtype=dtype,
        )

        self.batch_normalization1 = keras.layers.BatchNormalization(
            axis=channel_axis,
            epsilon=BN_EPSILON,
            momentum=BN_MOMENTUM,
            name=f"{name}_bn1",
            dtype=dtype,
        )

        self.activation1 = keras.layers.Activation(
            activation=activation,
            dtype=dtype,
        )

        self.pad = keras.layers.ZeroPadding2D(
            padding=(padding, padding),
            name=f"{name}_pad",
            dtype=dtype,
        )

        self.conv2 = keras.layers.Conv2D(
            expanded_channels,
            kernel_size,
            strides=stride,
            padding="valid",
            groups=expanded_channels,
            data_format=keras.config.image_data_format(),
            use_bias=False,
            name=f"{name}_conv2",
            dtype=dtype,
        )
        self.batch_normalization2 = keras.layers.BatchNormalization(
            axis=channel_axis,
            epsilon=BN_EPSILON,
            momentum=BN_MOMENTUM,
            name=f"{name}_bn2",
            dtype=dtype,
        )

        self.activation2 = keras.layers.Activation(
            activation=activation,
            dtype=dtype,
        )

        self.squeeze_excite = None
        if self.squeeze_excite_ratio:
            se_filters = expanded_channels
            self.squeeze_excite = SqueezeAndExcite2D(
                filters=se_filters,
                bottleneck_filters=adjust_channels(
                    se_filters * squeeze_excite_ratio
                ),
                squeeze_activation="relu",
                excite_activation=keras.activations.hard_sigmoid,
                name=f"{name}_se",
                dtype=dtype,
            )

        self.conv3 = keras.layers.Conv2D(
            filters,
            kernel_size=1,
            data_format=keras.config.image_data_format(),
            use_bias=False,
            name=f"{name}_conv3",
            dtype=dtype,
        )
        self.batch_normalization3 = keras.layers.BatchNormalization(
            axis=channel_axis,
            epsilon=BN_EPSILON,
            momentum=BN_MOMENTUM,
            name=f"{name}_bn3",
            dtype=dtype,
        )

    def build(self, input_shape):
        self.conv1.build(input_shape)
        input_shape = self.conv1.compute_output_shape(input_shape)
        self.batch_normalization1.build(input_shape)
        input_shape = self.batch_normalization1.compute_output_shape(
            input_shape
        )
        self.activation1.build(input_shape)
        input_shape = self.activation1.compute_output_shape(input_shape)
        self.pad.build(input_shape)
        input_shape = self.pad.compute_output_shape(input_shape)
        self.conv2.build(input_shape)
        input_shape = self.conv2.compute_output_shape(input_shape)
        self.batch_normalization2.build(input_shape)
        input_shape = self.batch_normalization2.compute_output_shape(
            input_shape
        )
        self.activation2.build(input_shape)
        input_shape = self.activation2.compute_output_shape(input_shape)
        if self.squeeze_excite_ratio:
            self.squeeze_excite.build(input_shape)
            input_shape = self.squeeze_excite.compute_output_shape(input_shape)
        self.conv3.build(input_shape)
        input_shape = self.conv3.compute_output_shape(input_shape)
        self.batch_normalization3.build(input_shape)
        self.built = True

    def call(self, inputs):
        x = self.conv1(inputs)
        x = self.batch_normalization1(x)
        x = self.activation1(x)
        x = self.pad(x)
        x = self.conv2(x)
        x = self.batch_normalization2(x)
        x = self.activation2(x)
        if self.squeeze_excite:
            x = self.squeeze_excite(x)
        x = self.conv3(x)
        x = self.batch_normalization3(x)
        if self.stride == 1 and self.infilters == self.filters:
            x = inputs + x
        return x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "expansion": self.expansion,
                "infilters": self.infilters,
                "filters": self.filters,
                "kernel_size": self.kernel_size,
                "stride": self.stride,
                "squeeze_excite_ratio": self.squeeze_excite_ratio,
                "activation": self.activation,
                "padding": self.padding,
                "name": self.name,
            }
        )
        return config


class ConvBnActBlock(keras.layers.Layer):
    """
    A ConvBnActBlock consists of a convultion, batchnorm, and activation layer

    Args:
        filters: Integer, the dimensionality of the output space
            (i.e. the number of output filters in the pointwise convolution).
        activation: The activation function to apply to the signal at the end.
        dtype: `None` or str or `keras.mixed_precision.DTypePolicy`. The dtype
            to use for the model's computations and weights.

    Input shape (when called as a layer):
        4D tensor with shape: `(batch, rows, cols, channels)` in "channels_last"
        4D tensor with shape: `(batch, channels, rows, cols)` in
            "channels_first"

    Returns:
        Output tensor of block.
    """

    def __init__(
        self,
        filter,
        activation,
        name=None,
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.filter = filter
        self.activation = activation
        self.name = name

        channel_axis = (
            -1 if keras.config.image_data_format() == "channels_last" else 1
        )
        self.conv = keras.layers.Conv2D(
            filter,
            kernel_size=1,
            data_format=keras.config.image_data_format(),
            use_bias=False,
            name=f"{name}_conv",
            dtype=dtype,
        )
        self.batch_normalization = keras.layers.BatchNormalization(
            axis=channel_axis,
            epsilon=BN_EPSILON,
            momentum=BN_MOMENTUM,
            name=f"{name}_bn",
            dtype=dtype,
        )
        self.activation_layer = keras.layers.Activation(
            activation,
            dtype=dtype,
        )

    def build(self, input_shape):
        self.conv.build(input_shape)
        input_shape = self.conv.compute_output_shape(input_shape)
        self.batch_normalization.build(input_shape)
        input_shape = self.batch_normalization.compute_output_shape(input_shape)
        self.activation_layer.build(input_shape)
        self.built = True

    def call(self, inputs):
        x = self.conv(inputs)
        x = self.batch_normalization(x)
        x = self.activation_layer(x)
        return x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "filter": self.filter,
                "activation": self.activation,
                "name": self.name,
            }
        )
        return config


@keras_hub_export("keras_hub.models.MobileNetBackbone")
class MobileNetBackbone(Backbone):
    """Instantiates the MobileNet architecture.

    MobileNet is a lightweight convolutional neural network (CNN)
    optimized for mobile and edge devices, striking a balance between
    accuracy and efficiency. By employing depthwise separable convolutions
    and techniques like Squeeze-and-Excitation (SE) blocks,
    MobileNet models are highly suitable for real-time applications on
    resource-constrained devices.

    References:
        - [MobileNets: Efficient Convolutional Neural Networks
       for Mobile Vision Applications](
        https://arxiv.org/abs/1704.04861)
        - [MobileNetV2: Inverted Residuals and Linear Bottlenecks](
        https://arxiv.org/abs/1801.04381) (CVPR 2018)
        - [Searching for MobileNetV3](https://arxiv.org/pdf/1905.02244.pdf)
        (ICCV 2019)

    Args:
        stackwise_expansion: list of list of ints, the expanded filters for
            each inverted residual block for each block in the model.
        stackwise_num_blocks: list of ints, number of inversted residual blocks
            per block
        stackwise_num_filters: list of list of ints, number of filters for
            each inverted residual block in the model.
        stackwise_kernel_size: list of list of ints, kernel size for each
            inverted residual block in the model.
        stackwise_num_strides: list of list of ints, stride length for each
            inverted residual block in the model.
        stackwise_se_ratio: se ratio for each inverted residual block in the
            model. 0 if dont want to add Squeeze and Excite layer.
        stackwise_activation: list of list of activation functions, for each
            inverted residual block in the model.
        stackwise_padding: list of list of int, to provide padding values for
            each inverted residual block in the model.
        output_num_filters: specifies whether to add conv and batch_norm in the
            end, if set to None, it will not add these layers in the end.
            'None' for MobileNetV1
        depthwise_filters: int, number of filters in depthwise separable
            convolution layer,
        last_layer_filter: int, channels/filters for the head ConvBnAct block
        squeeze_and_excite: float, squeeze and excite ratio in the depthwise
            layer, None, if dont want to do squeeze and excite
        image_shape: optional shape tuple, defaults to (224, 224, 3).
        input_activation: activation function to be used in the input layer
            'hard_swish' for MobileNetV3,
            'relu6' for MobileNetV1 and MobileNetV2
        output_activation: activation function to be used in the output layer
            'hard_swish' for MobileNetV3,
            'relu6' for MobileNetV1 and MobileNetV2
        input_num_filters: int, channels/filters for the input before the stem
            input_conv
        dtype: `None` or str or `keras.mixed_precision.DTypePolicy`. The dtype
            to use for the model's computations and weights.


    Example:
    ```python
    input_data = tf.ones(shape=(8, 224, 224, 3))

    # Randomly initialized backbone with a custom config
    model = MobileNetBackbone(
        stackwise_expansion=[
                [40, 56],
                [64, 144, 144],
                [72, 72],
                [144, 288, 288],
            ],
            stackwise_num_blocks=[2, 3, 2, 3],
            stackwise_num_filters=[
                [16, 16],
                [24, 24, 24],
                [24, 24],
                [48, 48, 48],
            ],
            stackwise_kernel_size=[[3, 3], [5, 5, 5], [5, 5], [5, 5, 5]],
            stackwise_num_strides=[[2, 1], [2, 1, 1], [1, 1], [2, 1, 1]],
            stackwise_se_ratio=[
                [None, None],
                [0.25, 0.25, 0.25],
                [0.3, 0.3],
                [0.3, 0.25, 0.25],
            ],
            stackwise_activation=[
                ["relu", "relu"],
                ["hard_swish", "hard_swish", "hard_swish"],
                ["hard_swish", "hard_swish"],
                ["hard_swish", "hard_swish", "hard_swish"],
            ],
            output_num_filters=288,
            input_activation="hard_swish",
            output_activation="hard_swish",
            input_num_filters=16,
            image_shape=(224, 224, 3),
            depthwise_filters=8,
            squeeze_and_excite=0.5,

    )
    output = model(input_data)
    ```
    """

    def __init__(
        self,
        stackwise_expansion,
        stackwise_num_blocks,
        stackwise_num_filters,
        stackwise_kernel_size,
        stackwise_num_strides,
        stackwise_se_ratio,
        stackwise_activation,
        stackwise_padding,
        output_num_filters,
        depthwise_filters,
        depthwise_stride,
        depthwise_residual,
        last_layer_filter,
        squeeze_and_excite=None,
        image_shape=(None, None, 3),
        input_activation="hard_swish",
        output_activation="hard_swish",
        input_num_filters=16,
        dtype=None,
        **kwargs,
    ):
        # === Functional Model ===
        channel_axis = (
            -1 if keras.config.image_data_format() == "channels_last" else 1
        )

        image_input = keras.layers.Input(shape=image_shape)
        x = image_input
        input_num_filters = adjust_channels(input_num_filters)

        x = keras.layers.ZeroPadding2D(
            padding=(1, 1),
            name="input_pad",
            dtype=dtype,
        )(x)
        x = keras.layers.Conv2D(
            input_num_filters,
            kernel_size=3,
            strides=(2, 2),
            data_format=keras.config.image_data_format(),
            use_bias=False,
            name="input_conv",
            dtype=dtype,
        )(x)
        x = keras.layers.BatchNormalization(
            axis=channel_axis,
            epsilon=BN_EPSILON,
            momentum=BN_MOMENTUM,
            name="input_batch_norm",
            dtype=dtype,
        )(x)
        x = keras.layers.Activation(
            input_activation,
            dtype=dtype,
        )(x)

        x = DepthwiseConvBlock(
            input_num_filters,
            depthwise_filters,
            stride=depthwise_stride,
            squeeze_excite_ratio=squeeze_and_excite,
            residual=depthwise_residual,
            name="block_0",
            dtype=dtype,
        )(x)

        for block in range(len(stackwise_num_blocks)):
            for inverted_block in range(stackwise_num_blocks[block]):
                infilters = x.shape[channel_axis]
                x = InvertedResidualBlock(
                    expansion=stackwise_expansion[block][inverted_block],
                    infilters=infilters,
                    filters=adjust_channels(
                        stackwise_num_filters[block][inverted_block]
                    ),
                    kernel_size=stackwise_kernel_size[block][inverted_block],
                    stride=stackwise_num_strides[block][inverted_block],
                    squeeze_excite_ratio=stackwise_se_ratio[block][
                        inverted_block
                    ],
                    activation=stackwise_activation[block][inverted_block],
                    padding=stackwise_padding[block][inverted_block],
                    name=f"block_{block + 1}_{inverted_block}",
                    dtype=dtype,
                )(x)

        x = ConvBnActBlock(
            filter=adjust_channels(last_layer_filter),
            activation="hard_swish",
            name=f"block_{len(stackwise_num_blocks) + 1}_0",
            dtype=dtype,
        )(x)

        super().__init__(inputs=image_input, outputs=x, dtype=dtype, **kwargs)

        # === Config ===
        self.stackwise_expansion = stackwise_expansion
        self.stackwise_num_blocks = stackwise_num_blocks
        self.stackwise_num_filters = stackwise_num_filters
        self.stackwise_kernel_size = stackwise_kernel_size
        self.stackwise_num_strides = stackwise_num_strides
        self.stackwise_se_ratio = stackwise_se_ratio
        self.stackwise_activation = stackwise_activation
        self.stackwise_padding = stackwise_padding
        self.input_num_filters = input_num_filters
        self.output_num_filters = output_num_filters
        self.depthwise_filters = depthwise_filters
        self.depthwise_stride = depthwise_stride
        self.depthwise_residual = depthwise_residual
        self.last_layer_filter = last_layer_filter
        self.squeeze_and_excite = squeeze_and_excite
        self.input_activation = input_activation
        self.output_activation = output_activation
        self.image_shape = image_shape

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "stackwise_expansion": self.stackwise_expansion,
                "stackwise_num_blocks": self.stackwise_num_blocks,
                "stackwise_num_filters": self.stackwise_num_filters,
                "stackwise_kernel_size": self.stackwise_kernel_size,
                "stackwise_num_strides": self.stackwise_num_strides,
                "stackwise_se_ratio": self.stackwise_se_ratio,
                "stackwise_activation": self.stackwise_activation,
                "stackwise_padding": self.stackwise_padding,
                "image_shape": self.image_shape,
                "input_num_filters": self.input_num_filters,
                "output_num_filters": self.output_num_filters,
                "depthwise_filters": self.depthwise_filters,
                "depthwise_stride": self.depthwise_stride,
                "depthwise_residual": self.depthwise_residual,
                "last_layer_filter": self.last_layer_filter,
                "squeeze_and_excite": self.squeeze_and_excite,
                "input_activation": self.input_activation,
                "output_activation": self.output_activation,
            }
        )
        return config
